@@ -1,11 +1,11 @@
 /*
- * "$Id: http.c,v 1.57 1999/12/21 02:26:45 mike Exp $"
+ * "$Id: http.c,v 1.94 2002/02/12 19:05:38 mike Exp $"
  *
  *   HTTP routines for the Common UNIX Printing System (CUPS) scheduler.
  *
- *   Copyright 1997-1999 by Easy Software Products, all rights reserved.
+ *   Copyright 1997-2002 by Easy Software Products, all rights reserved.
  *
- *   These statusd instructions, statements, and computer programs are the
+ *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
  *   copyright law.  Distribution and use rights are outlined in the file
  *   "LICENSE.txt" which should have been included with this file.  If this
@@ -23,39 +23,45 @@
  *
  * Contents:
  *
- *   httpInitialize()    - Initialize the HTTP interface library and set the
- *                         default HTTP proxy (if any).
- *   httpCheck()         - Check to see if there is a pending response from
- *                         the server.
- *   httpClose()         - Close an HTTP connection...
- *   httpConnect()       - Connect to a HTTP server.
- *   httpReconnect()     - Reconnect to a HTTP server...
- *   httpSeparate()      - Separate a Universal Resource Identifier into its
- *                         components.
- *   httpSetField()      - Set the value of an HTTP header.
- *   httpDelete()        - Send a DELETE request to the server.
- *   httpGet()           - Send a GET request to the server.
- *   httpHead()          - Send a HEAD request to the server.
- *   httpOptions()       - Send an OPTIONS request to the server.
- *   httpPost()          - Send a POST request to the server.
- *   httpPut()           - Send a PUT request to the server.
- *   httpTrace()         - Send an TRACE request to the server.
- *   httpFlush()         - Flush data from a HTTP connection.
- *   httpRead()          - Read data from a HTTP connection.
- *   httpWrite()         - Write data to a HTTP connection.
- *   httpGets()          - Get a line of text from a HTTP connection.
- *   httpPrintf()        - Print a formatted string to a HTTP connection.
- *   httpStatus()        - Return a short string describing a HTTP status code.
- *   httpGetDateString() - Get a formatted date/time string from a time value.
- *   httpGetDateTime()   - Get a time value from a formatted date/time string.
- *   httpUpdate()        - Update the current HTTP state for incoming data.
- *   httpDecode64()      - Base64-decode a string.
- *   httpEncode64()      - Base64-encode a string.
- *   httpGetLength()     - Get the amount of data remaining from the
- *                         content-length or transfer-encoding fields.
- *   http_field()        - Return the field index for a field name.
- *   http_send()         - Send a request with all fields and the trailing
- *                         blank line.
+ *   httpInitialize()     - Initialize the HTTP interface library and set the
+ *                          default HTTP proxy (if any).
+ *   httpCheck()          - Check to see if there is a pending response from
+ *                          the server.
+ *   httpClose()          - Close an HTTP connection...
+ *   httpConnect()        - Connect to a HTTP server.
+ *   httpConnectEncrypt() - Connect to a HTTP server using encryption.
+ *   httpEncryption()     - Set the required encryption on the link.
+ *   httpReconnect()      - Reconnect to a HTTP server...
+ *   httpGetHostByName()  - Lookup a hostname or IP address, and return
+ *                          address records for the specified name.
+ *   httpSeparate()       - Separate a Universal Resource Identifier into its
+ *                          components.
+ *   httpGetSubField()    - Get a sub-field value.
+ *   httpSetField()       - Set the value of an HTTP header.
+ *   httpDelete()         - Send a DELETE request to the server.
+ *   httpGet()            - Send a GET request to the server.
+ *   httpHead()           - Send a HEAD request to the server.
+ *   httpOptions()        - Send an OPTIONS request to the server.
+ *   httpPost()           - Send a POST request to the server.
+ *   httpPut()            - Send a PUT request to the server.
+ *   httpTrace()          - Send an TRACE request to the server.
+ *   httpFlush()          - Flush data from a HTTP connection.
+ *   httpRead()           - Read data from a HTTP connection.
+ *   httpWrite()          - Write data to a HTTP connection.
+ *   httpGets()           - Get a line of text from a HTTP connection.
+ *   httpPrintf()         - Print a formatted string to a HTTP connection.
+ *   httpStatus()         - Return a short string describing a HTTP status code.
+ *   httpGetDateString()  - Get a formatted date/time string from a time value.
+ *   httpGetDateTime()    - Get a time value from a formatted date/time string.
+ *   httpUpdate()         - Update the current HTTP state for incoming data.
+ *   httpDecode64()       - Base64-decode a string.
+ *   httpEncode64()       - Base64-encode a string.
+ *   httpGetLength()      - Get the amount of data remaining from the
+ *                          content-length or transfer-encoding fields.
+ *   http_field()         - Return the field index for a field name.
+ *   http_send()          - Send a request with all fields and the trailing
+ *                          blank line.
+ *   http_upgrade()       - Force upgrade to TLS encryption.
  */
 
 /*
@@ -74,9 +80,16 @@
 #include "ipp.h"
 #include "debug.h"
 
-#if !defined(WIN32) && !defined(__EMX__)
+#ifndef WIN32
 #  include <signal.h>
-#endif /* !WIN32 && !__EMX__ */
+#endif /* !WIN32 */
+
+#ifdef HAVE_LIBSSL
+#  include <openssl/err.h>
+#  include <openssl/rand.h>
+#  include <openssl/ssl.h>
+#endif /* HAVE_LIBSSL */
+
 
 /*
  * Some operating systems have done away with the Fxxxx constants for
@@ -95,6 +108,9 @@
 static http_field_t	http_field(const char *name);
 static int		http_send(http_t *http, http_state_t request,
 			          const char *uri);
+#ifdef HAVE_LIBSSL
+static int		http_upgrade(http_t *http);
+#endif /* HAVE_LIBSSL */
 
 
 /*
@@ -166,7 +182,13 @@ static const char	*months[12] =
 void
 httpInitialize(void)
 {
-#if defined(WIN32) || defined(__EMX__)
+#ifdef HAVE_LIBSSL
+  struct timeval	curtime;	/* Current time in microseconds */
+  int			i;		/* Looping var */
+  unsigned char		data[1024];	/* Seed data */
+#endif /* HAVE_LIBSSL */
+
+#ifdef WIN32
   WSADATA	winsockdata;	/* WinSock data */
   static int	initialized = 0;/* Has WinSock been initialized? */
 
@@ -188,7 +210,25 @@ httpInitialize(void)
   sigaction(SIGPIPE, &action, NULL);
 #else
   signal(SIGPIPE, SIG_IGN);
-#endif /* WIN32 || __EMX__ */
+#endif /* WIN32 */
+
+#ifdef HAVE_LIBSSL
+  SSL_load_error_strings();
+  SSL_library_init();
+
+ /*
+  * Using the current time is a dubious random seed, but on some systems
+  * it is the best we can do (on others, this seed isn't even used...)
+  */
+
+  gettimeofday(&curtime, NULL);
+  srand(curtime.tv_sec + curtime.tv_usec);
+
+  for (i = 0; i < sizeof(data); i ++)
+    data[i] = rand(); /* Yes, this is a poor source of random data... */
+
+  RAND_seed(&data, sizeof(data));
+#endif /* HAVE_LIBSSL */
 }
 
 
@@ -234,9 +274,29 @@ httpCheck(http_t *http)		/* I - HTTP connection */
 void
 httpClose(http_t *http)		/* I - Connection to close */
 {
-  if (http == NULL)
+#ifdef HAVE_LIBSSL
+  SSL_CTX	*context;	/* Context for encryption */
+  SSL		*conn;		/* Connection for encryption */
+#endif /* HAVE_LIBSSL */
+
+
+  if (!http)
     return;
 
+#ifdef HAVE_LIBSSL
+  if (http->tls)
+  {
+    conn    = (SSL *)(http->tls);
+    context = SSL_get_SSL_CTX(conn);
+
+    SSL_shutdown(conn);
+    SSL_CTX_free(context);
+    SSL_free(conn);
+
+    http->tls = NULL;
+  }
+#endif /* HAVE_LIBSSL */
+
 #ifdef WIN32
   closesocket(http->fd);
 #else
@@ -255,17 +315,64 @@ http_t *			/* O - New HTTP connection */
 httpConnect(const char *host,	/* I - Host to connect to */
             int        port)	/* I - Port number */
 {
+  http_encryption_t	encrypt;/* Type of encryption to use */
+
+
+ /*
+  * Set the default encryption status...
+  */
+
+  if (port == 443)
+    encrypt = HTTP_ENCRYPT_ALWAYS;
+  else
+    encrypt = HTTP_ENCRYPT_IF_REQUESTED;
+
+  return (httpConnectEncrypt(host, port, encrypt));
+}
+
+
+/*
+ * 'httpConnectEncrypt()' - Connect to a HTTP server using encryption.
+ */
+
+http_t *				/* O - New HTTP connection */
+httpConnectEncrypt(const char *host,	/* I - Host to connect to */
+                   int        port,	/* I - Port number */
+		   http_encryption_t encrypt)
+					/* I - Type of encryption to use */
+{
+  int			i;		/* Looping var */
   http_t		*http;		/* New HTTP connection */
   struct hostent	*hostaddr;	/* Host address data */
 
 
+  if (host == NULL)
+    return (NULL);
+
   httpInitialize();
 
  /*
   * Lookup the host...
   */
 
-  if ((hostaddr = gethostbyname(host)) == NULL)
+  if ((hostaddr = httpGetHostByName(host)) == NULL)
+  {
+   /*
+    * This hack to make users that don't have a localhost entry in
+    * their hosts file or DNS happy...
+    */
+
+    if (strcasecmp(host, "localhost") != 0)
+      return (NULL);
+    else if ((hostaddr = httpGetHostByName("127.0.0.1")) == NULL)
+      return (NULL);
+  }
+
+ /*
+  * Verify that it is an IPv4 address (IPv6 support will come in CUPS 1.2...)
+  */
+
+  if (hostaddr->h_addrtype != AF_INET || hostaddr->h_length != 4)
     return (NULL);
 
  /*
@@ -279,26 +386,88 @@ httpConnect(const char *host,	/* I - Host to connect to */
   http->version  = HTTP_1_1;
   http->blocking = 1;
   http->activity = time(NULL);
+  http->fd       = -1;
 
  /*
   * Copy the hostname and port and then "reconnect"...
   */
 
   strncpy(http->hostname, host, sizeof(http->hostname) - 1);
-  memcpy((char *)&(http->hostaddr.sin_addr), hostaddr->h_addr, hostaddr->h_length);
   http->hostaddr.sin_family = hostaddr->h_addrtype;
 #ifdef WIN32
   http->hostaddr.sin_port   = htons((u_short)port);
 #else
   http->hostaddr.sin_port   = htons(port);
 #endif /* WIN32 */
-  if (httpReconnect(http))
-  {
-    free(http);
-    return (NULL);
-  }
+
+ /*
+  * Set the encryption status...
+  */
+
+  if (port == 443)	/* Always use encryption for https */
+    http->encryption = HTTP_ENCRYPT_ALWAYS;
   else
-    return (http);
+    http->encryption = encrypt;
+
+ /*
+  * Loop through the addresses we have until one of them connects...
+  */
+
+  strncpy(http->hostname, host, sizeof(http->hostname) - 1);
+
+  for (i = 0; hostaddr->h_addr_list[i]; i ++)
+  {
+   /*
+    * Load the address...
+    */
+
+    memcpy((char *)&(http->hostaddr.sin_addr), hostaddr->h_addr_list[i],
+           hostaddr->h_length);
+
+   /*
+    * Connect to the remote system...
+    */
+
+    if (!httpReconnect(http))
+      return (http);
+  }
+
+ /*
+  * Could not connect to any known address - bail out!
+  */
+
+  free(http);
+  return (NULL);
+}
+
+
+/*
+ * 'httpEncryption()' - Set the required encryption on the link.
+ */
+
+int					/* O - -1 on error, 0 on success */
+httpEncryption(http_t            *http,	/* I - HTTP data */
+               http_encryption_t e)	/* I - New encryption preference */
+{
+#ifdef HAVE_LIBSSL
+  if (!http)
+    return (0);
+
+  http->encryption = e;
+
+  if ((http->encryption == HTTP_ENCRYPT_ALWAYS && !http->tls) ||
+      (http->encryption == HTTP_ENCRYPT_NEVER && http->tls))
+    return (httpReconnect(http));
+  else if (http->encryption == HTTP_ENCRYPT_REQUIRED && !http->tls)
+    return (http_upgrade(http));
+  else
+    return (0);
+#else
+  if (e == HTTP_ENCRYPT_ALWAYS || e == HTTP_ENCRYPT_REQUIRED)
+    return (-1);
+  else
+    return (0);
+#endif /* HAVE_LIBSSL */
 }
 
 
@@ -309,14 +478,30 @@ httpConnect(const char *host,	/* I - Host to connect to */
 int				/* O - 0 on success, non-zero on failure */
 httpReconnect(http_t *http)	/* I - HTTP data */
 {
-  int	val;			/* Socket option value */
+  int		val;		/* Socket option value */
+#ifdef HAVE_LIBSSL
+  SSL_CTX	*context;	/* Context for encryption */
+  SSL		*conn;		/* Connection for encryption */
 
 
+  if (http->tls)
+  {
+    conn    = (SSL *)(http->tls);
+    context = SSL_get_SSL_CTX(conn);
+
+    SSL_shutdown(conn);
+    SSL_CTX_free(context);
+    SSL_free(conn);
+
+    http->tls = NULL;
+  }
+#endif /* HAVE_LIBSSL */
+
  /*
   * Close any previously open socket...
   */
 
-  if (http->fd)
+  if (http->fd >= 0)
 #ifdef WIN32
     closesocket(http->fd);
 #else
@@ -329,11 +514,11 @@ httpReconnect(http_t *http)	/* I - HTTP data */
 
   if ((http->fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
   {
-#if defined(WIN32) || defined(__EMX__)
+#ifdef WIN32
     http->error  = WSAGetLastError();
 #else
     http->error  = errno;
-#endif /* WIN32 || __EMX__ */
+#endif /* WIN32 */
     http->status = HTTP_ERROR;
     return (-1);
   }
@@ -351,6 +536,16 @@ httpReconnect(http_t *http)	/* I - HTTP data */
   setsockopt(http->fd, SOL_SOCKET, SO_REUSEPORT, &val, sizeof(val));
 #endif /* SO_REUSEPORT */
 
+ /*
+  * Using TCP_NODELAY improves responsiveness, especially on systems
+  * with a slow loopback interface...  Since we write large buffers
+  * when sending print files and requests, there shouldn't be any
+  * performance penalty for this...
+  */
+
+  val = 1;
+  setsockopt(http->fd, IPPROTO_TCP, TCP_NODELAY, &val, sizeof(val)); 
+
  /*
   * Connect to the server...
   */
@@ -358,11 +553,11 @@ httpReconnect(http_t *http)	/* I - HTTP data */
   if (connect(http->fd, (struct sockaddr *)&(http->hostaddr),
               sizeof(http->hostaddr)) < 0)
   {
-#if defined(WIN32) || defined(__EMX__)
+#ifdef WIN32
     http->error  = WSAGetLastError();
 #else
     http->error  = errno;
-#endif /* WIN32 || __EMX__ */
+#endif /* WIN32 */
     http->status = HTTP_ERROR;
 
 #ifdef WIN32
@@ -371,16 +566,125 @@ httpReconnect(http_t *http)	/* I - HTTP data */
     close(http->fd);
 #endif
 
+    http->fd = -1;
+
     return (-1);
   }
 
   http->error  = 0;
   http->status = HTTP_CONTINUE;
 
+#ifdef HAVE_LIBSSL
+  if (http->encryption == HTTP_ENCRYPT_ALWAYS)
+  {
+   /*
+    * Always do encryption via SSL.
+    */
+
+    context = SSL_CTX_new(SSLv23_method());
+    conn    = SSL_new(context);
+
+    SSL_set_fd(conn, http->fd);
+    if (SSL_connect(conn) != 1)
+    {
+      SSL_CTX_free(context);
+      SSL_free(conn);
+
+#ifdef WIN32
+      http->error  = WSAGetLastError();
+#else
+      http->error  = errno;
+#endif /* WIN32 */
+      http->status = HTTP_ERROR;
+
+#ifdef WIN32
+      closesocket(http->fd);
+#else
+      close(http->fd);
+#endif
+
+      return (-1);
+    }
+
+    http->tls = conn;
+  }
+  else if (http->encryption == HTTP_ENCRYPT_REQUIRED)
+    return (http_upgrade(http));
+#endif /* HAVE_LIBSSL */
+
   return (0);
 }
 
 
+/*
+ * 'httpGetHostByName()' - Lookup a hostname or IP address, and return
+ *                         address records for the specified name.
+ */
+
+struct hostent *			/* O - Host entry */
+httpGetHostByName(const char *name)	/* I - Hostname or IP address */
+{
+  unsigned		ip[4];		/* IP address components */
+  static unsigned	packed_ip;	/* Packed IPv4 address */
+  static char		*packed_ptr[2];	/* Pointer to packed address */
+  static struct hostent	host_ip;	/* Host entry for IP address */
+
+#if defined(__APPLE__)
+  /* OS X hack to avoid it's ocassional long delay in lookupd */
+  static char sLoopback[] = "127.0.0.1";
+  if (strcmp(name, "localhost") == 0)
+    name = sLoopback;
+#endif /* __APPLE__ */
+
+ /*
+  * This function is needed because some operating systems have a
+  * buggy implementation of httpGetHostByName() that does not support
+  * IP addresses.  If the first character of the name string is a
+  * number, then sscanf() is used to extract the IP components.
+  * We then pack the components into an IPv4 address manually,
+  * since the inet_aton() function is deprecated.  We use the
+  * htonl() macro to get the right byte order for the address.
+  */
+
+  if (isdigit(name[0]))
+  {
+   /*
+    * We have an IP address; break it up and provide the host entry
+    * to the caller.  Currently only supports IPv4 addresses, although
+    * it should be trivial to support IPv6 in CUPS 1.2.
+    */
+
+    if (sscanf(name, "%u.%u.%u.%u", ip, ip + 1, ip + 2, ip + 3) != 4)
+      return (NULL); /* Must have 4 numbers */
+
+    packed_ip = htonl(((((((ip[0] << 8) | ip[1]) << 8) | ip[2]) << 8) | ip[3]));
+
+   /*
+    * Fill in the host entry and return it...
+    */
+
+    host_ip.h_name      = (char *)name;
+    host_ip.h_aliases   = NULL;
+    host_ip.h_addrtype  = AF_INET;
+    host_ip.h_length    = 4;
+    host_ip.h_addr_list = packed_ptr;
+    packed_ptr[0]       = (char *)(&packed_ip);
+    packed_ptr[1]       = NULL;
+
+    return (&host_ip);
+  }
+  else
+  {
+   /*
+    * Use the gethostbyname() function to get the IP address for
+    * the name...
+    */
+
+    return (gethostbyname(name));
+  }
+}
+
+
 /*
  * 'httpSeparate()' - Separate a Universal Resource Identifier into its
  *                    components.
@@ -394,66 +698,98 @@ httpSeparate(const char *uri,		/* I - Universal Resource Identifier */
 	     int        *port,		/* O - Port number to use */
              char       *resource)	/* O - Resource/filename [1024] */
 {
-  char	*ptr;			/* Pointer into string... */
+  char		*ptr;				/* Pointer into string... */
+  const char	*atsign,			/* @ sign */
+		*slash;				/* Separator */
+  char		safeuri[HTTP_MAX_URI];		/* "Safe" local copy of URI */
 
 
+ /*
+  * Range check input...
+  */
+
   if (uri == NULL || method == NULL || username == NULL || host == NULL ||
       port == NULL || resource == NULL)
     return;
 
+ /*
+  * Copy the URL to a local string to make sure we don't have a URL
+  * longer than HTTP_MAX_URI characters long...
+  */
+
+  strncpy(safeuri, uri, sizeof(safeuri));
+  safeuri[sizeof(safeuri) - 1] = '\0';
+
+  uri = safeuri;
+
  /*
   * Grab the method portion of the URI...
   */
 
-  ptr = host;
-  while (*uri != ':' && *uri != '\0')
-    *ptr++ = *uri++;
-
-  *ptr = '\0';
-  if (*uri == ':')
-    uri ++;
-
- /*
-  * If the method contains a period or slash, then it's probably
-  * hostname/filename...
-  */
-
-  if (strchr(host, '.') != NULL || strchr(host, '/') != NULL || *uri == '\0')
+  if (strncmp(uri, "//", 2) == 0)
   {
-    if ((ptr = strchr(host, '/')) != NULL)
-    {
-      strncpy(resource, ptr, HTTP_MAX_URI - 1);
-      resource[HTTP_MAX_URI - 1] = '\0';
-      *ptr = '\0';
-    }
-    else
-      resource[0] = '\0';
+   /*
+    * Workaround for HP IPP client bug...
+    */
 
-    if (isdigit(*uri))
-    {
-     /*
-      * OK, we have "hostname:port[/resource]"...
-      */
-
-      *port = strtol(uri, (char **)&uri, 10);
-
-      if (*uri == '/')
-      {
-        strncpy(resource, uri, HTTP_MAX_URI - 1);
-        resource[HTTP_MAX_URI - 1] = '\0';
-      }
-    }
-    else
-      *port = 0;
-
-    strcpy(method, "http");
-    username[0] = '\0';
-    return;
+    strcpy(method, "ipp");
   }
   else
   {
-    strncpy(method, host, 31);
-    method[31] = '\0';
+   /*
+    * Standard URI with method...
+    */
+
+    for (ptr = host; *uri != ':' && *uri != '\0'; uri ++)
+      if (ptr < (host + HTTP_MAX_URI - 1))
+        *ptr++ = *uri;
+
+    *ptr = '\0';
+    if (*uri == ':')
+      uri ++;
+
+   /*
+    * If the method contains a period or slash, then it's probably
+    * hostname/filename...
+    */
+
+    if (strchr(host, '.') != NULL || strchr(host, '/') != NULL || *uri == '\0')
+    {
+      if ((ptr = strchr(host, '/')) != NULL)
+      {
+	strncpy(resource, ptr, HTTP_MAX_URI);
+	resource[HTTP_MAX_URI - 1] = '\0';
+	*ptr = '\0';
+      }
+      else
+	resource[0] = '\0';
+
+      if (isdigit(*uri))
+      {
+       /*
+	* OK, we have "hostname:port[/resource]"...
+	*/
+
+	*port = strtol(uri, (char **)&uri, 10);
+
+	if (*uri == '/')
+	{
+          strncpy(resource, uri, HTTP_MAX_URI);
+          resource[HTTP_MAX_URI - 1] = '\0';
+	}
+      }
+      else
+	*port = 631;
+
+      strcpy(method, "http");
+      username[0] = '\0';
+      return;
+    }
+    else
+    {
+      strncpy(method, host, 31);
+      method[31] = '\0';
+    }
   }
 
  /*
@@ -472,49 +808,56 @@ httpSeparate(const char *uri,		/* I - Universal Resource Identifier */
   }
 
  /*
-  * Grab the hostname...
+  * Grab the username, if any...
   */
 
   while (*uri == '/')
     uri ++;
 
-  ptr = host;
-  while (!(*uri == ':' && isdigit(uri[1])) && *uri != '@' && *uri != '/' && *uri != '\0')
-    *ptr ++ = *uri ++;
+  if ((slash = strchr(uri, '/')) == NULL)
+    slash = uri + strlen(uri);
 
-  *ptr = '\0';
-
-  if (*uri == '@')
+  if ((atsign = strchr(uri, '@')) != NULL && atsign < slash)
   {
    /*
-    * Got a username...
+    * Got a username:password combo...
     */
 
-    strncpy(username, host, 31);
-    username[31] = '\0';
-    
-    ptr = host;
-    uri ++;
-    while (*uri != ':' && *uri != '/' && *uri != '\0')
-      *ptr ++ = *uri ++;
+    for (ptr = username; uri < atsign; uri ++)
+      if (ptr < (username + HTTP_MAX_URI - 1))
+	*ptr++ = *uri;
 
     *ptr = '\0';
+
+    uri = atsign + 1;
   }
   else
     username[0] = '\0';
 
-  if (*uri == '\0')
-  {
-   /*
-    * Hostname but no port or path...
-    */
+ /*
+  * Grab the hostname...
+  */
 
-    *port = 0;
-    resource[0] = '/';
-    resource[1] = '\0';
-    return;
+  for (ptr = host; *uri != ':' && *uri != '/' && *uri != '\0'; uri ++)
+    if (ptr < (host + HTTP_MAX_URI - 1))
+      *ptr++ = *uri;
+
+  *ptr = '\0';
+
+  if (*uri != ':')
+  {
+    if (strcasecmp(method, "http") == 0)
+      *port = 80;
+    else if (strcasecmp(method, "https") == 0)
+      *port = 443;
+    else if (strcasecmp(method, "ipp") == 0)
+      *port = ippPort();
+    else if (strcasecmp(method, "socket") == 0)	/* Not registered yet... */
+      *port = 9100;
+    else
+      *port = 0;
   }
-  else if (*uri == ':')
+  else
   {
    /*
     * Parse port number...
@@ -528,33 +871,150 @@ httpSeparate(const char *uri,		/* I - Universal Resource Identifier */
       uri ++;
     }
   }
-  else
+
+  if (*uri == '\0')
   {
    /*
-    * Figure out the default port number based on the method...
+    * Hostname but no port or path...
     */
 
-    if (strcasecmp(method, "http") == 0)
-      *port = 80;
-    else if (strcasecmp(method, "https") == 0)
-      *port = 443;
-    else if (strcasecmp(method, "ipp") == 0)	/* Not registered yet... */
-      *port = ippPort();
-    else if (strcasecmp(method, "socket") == 0)	/* Not registered yet... */
-      *port = 9100;
-    else
-      *port = 0;
+    resource[0] = '/';
+    resource[1] = '\0';
+    return;
   }
 
  /*
   * The remaining portion is the resource string...
   */
 
-  strncpy(resource, uri, HTTP_MAX_URI - 1);
+  strncpy(resource, uri, HTTP_MAX_URI);
   resource[HTTP_MAX_URI - 1] = '\0';
 }
 
 
+/*
+ * 'httpGetSubField()' - Get a sub-field value.
+ */
+
+char *					/* O - Value or NULL */
+httpGetSubField(http_t       *http,	/* I - HTTP data */
+                http_field_t field,	/* I - Field index */
+                const char   *name,	/* I - Name of sub-field */
+		char         *value)	/* O - Value string */
+{
+  const char	*fptr;			/* Pointer into field */
+  char		temp[HTTP_MAX_VALUE],	/* Temporary buffer for name */
+		*ptr;			/* Pointer into string buffer */
+
+
+  if (http == NULL ||
+      field < HTTP_FIELD_ACCEPT_LANGUAGE ||
+      field > HTTP_FIELD_WWW_AUTHENTICATE ||
+      name == NULL || value == NULL)
+    return (NULL);
+
+  DEBUG_printf(("httpGetSubField(%p, %d, \"%s\", %p)\n",
+                http, field, name, value));
+
+  for (fptr = http->fields[field]; *fptr;)
+  {
+   /*
+    * Skip leading whitespace...
+    */
+
+    while (isspace(*fptr))
+      fptr ++;
+
+    if (*fptr == ',')
+    {
+      fptr ++;
+      continue;
+    }
+
+   /*
+    * Get the sub-field name...
+    */
+
+    for (ptr = temp;
+         *fptr && *fptr != '=' && !isspace(*fptr) && ptr < (temp + sizeof(temp) - 1);
+         *ptr++ = *fptr++);
+
+    *ptr = '\0';
+
+    DEBUG_printf(("name = \"%s\"\n", temp));
+
+   /*
+    * Skip trailing chars up to the '='...
+    */
+
+    while (isspace(*fptr))
+      fptr ++;
+
+    if (!*fptr)
+      break;
+
+    if (*fptr != '=')
+      continue;
+
+   /*
+    * Skip = and leading whitespace...
+    */
+
+    fptr ++;
+
+    while (isspace(*fptr))
+      fptr ++;
+
+    if (*fptr == '\"')
+    {
+     /*
+      * Read quoted string...
+      */
+
+      for (ptr = value, fptr ++;
+           *fptr && *fptr != '\"' && ptr < (value + HTTP_MAX_VALUE - 1);
+	   *ptr++ = *fptr++);
+
+      *ptr = '\0';
+
+      while (*fptr && *fptr != '\"')
+        fptr ++;
+
+      if (*fptr)
+        fptr ++;
+    }
+    else
+    {
+     /*
+      * Read unquoted string...
+      */
+
+      for (ptr = value;
+           *fptr && !isspace(*fptr) && *fptr != ',' && ptr < (value + HTTP_MAX_VALUE - 1);
+	   *ptr++ = *fptr++);
+
+      *ptr = '\0';
+
+      while (*fptr && !isspace(*fptr) && *fptr != ',')
+        fptr ++;
+    }
+
+    DEBUG_printf(("value = \"%s\"\n", value));
+
+   /*
+    * See if this is the one...
+    */
+
+    if (strcmp(name, temp) == 0)
+      return (value);
+  }
+
+  value[0] = '\0';
+
+  return (NULL);
+}
+
+
 /*
  * 'httpSetField()' - Set the value of an HTTP header.
  */
@@ -564,6 +1024,12 @@ httpSetField(http_t       *http,	/* I - HTTP data */
              http_field_t field,	/* I - Field index */
 	     const char   *value)	/* I - Value */
 {
+  if (http == NULL ||
+      field < HTTP_FIELD_ACCEPT_LANGUAGE ||
+      field > HTTP_FIELD_WWW_AUTHENTICATE ||
+      value == NULL)
+    return;
+
   strncpy(http->fields[field], value, HTTP_MAX_VALUE - 1);
   http->fields[field][HTTP_MAX_VALUE - 1] = '\0';
 }
@@ -730,6 +1196,47 @@ httpRead(http_t *http,			/* I - HTTP data */
   else if (length > http->data_remaining)
     length = http->data_remaining;
 
+  if (http->used == 0 && length <= 256)
+  {
+   /*
+    * Buffer small reads for better performance...
+    */
+
+    if (http->data_remaining > sizeof(http->buffer))
+      bytes = sizeof(http->buffer);
+    else
+      bytes = http->data_remaining;
+
+#ifdef HAVE_LIBSSL
+    if (http->tls)
+      bytes = SSL_read((SSL *)(http->tls), http->buffer, bytes);
+    else
+#endif /* HAVE_LIBSSL */
+    {
+      DEBUG_printf(("httpRead: reading %d bytes from socket into buffer...\n",
+                    bytes));
+
+      bytes = recv(http->fd, http->buffer, bytes, 0);
+
+      DEBUG_printf(("httpRead: read %d bytes from socket into buffer...\n",
+                    bytes));
+    }
+
+    if (bytes > 0)
+      http->used = bytes;
+    else if (bytes < 0)
+    {
+#ifdef WIN32
+      http->error = WSAGetLastError();
+#else
+      http->error = errno;
+#endif /* WIN32 */
+      return (-1);
+    }
+    else
+      return (0);
+  }
+
   if (http->used > 0)
   {
     if (length > http->used)
@@ -745,6 +1252,10 @@ httpRead(http_t *http,			/* I - HTTP data */
     if (http->used > 0)
       memcpy(http->buffer, http->buffer + length, http->used);
   }
+#ifdef HAVE_LIBSSL
+  else if (http->tls)
+    bytes = SSL_read((SSL *)(http->tls), buffer, length);
+#endif /* HAVE_LIBSSL */
   else
   {
     DEBUG_printf(("httpRead: reading %d bytes from socket...\n", length));
@@ -755,11 +1266,11 @@ httpRead(http_t *http,			/* I - HTTP data */
   if (bytes > 0)
     http->data_remaining -= bytes;
   else if (bytes < 0)
-#if defined(WIN32) || defined(__EMX__)
+#ifdef WIN32
     http->error = WSAGetLastError();
 #else
     http->error = errno;
-#endif /* WIN32 || __EMX__ */
+#endif /* WIN32 */
 
   if (http->data_remaining == 0)
   {
@@ -813,6 +1324,8 @@ httpWrite(http_t     *http,		/* I - HTTP data */
 
       if (http->state == HTTP_POST_RECV)
 	http->state ++;
+      else if (http->state == HTTP_PUT_RECV)
+        http->state = HTTP_STATUS;
       else
 	http->state = HTTP_WAITING;
 
@@ -827,7 +1340,13 @@ httpWrite(http_t     *http,		/* I - HTTP data */
 
   while (length > 0)
   {
+#ifdef HAVE_LIBSSL
+    if (http->tls)
+      bytes = SSL_write((SSL *)(http->tls), buffer, length);
+    else
+#endif /* HAVE_LIBSSL */
     bytes = send(http->fd, buffer, length, 0);
+
     if (bytes < 0)
     {
       DEBUG_puts("httpWrite: error writing data...\n");
@@ -890,11 +1409,11 @@ httpGets(char   *line,			/* I - Line to read into */
   * Pre-scan the buffer and see if there is a newline in there...
   */
 
-#if defined(WIN32) || defined(__EMX__)
+#ifdef WIN32
   WSASetLastError(0);
 #else
   errno = 0;
-#endif /* WIN32 || __EMX__ */
+#endif /* WIN32 */
 
   do
   {
@@ -907,19 +1426,27 @@ httpGets(char   *line,			/* I - Line to read into */
       else
 	bufptr ++;
 
-    if (bufptr >= bufend)
+    if (bufptr >= bufend && http->used < HTTP_MAX_BUFFER)
     {
      /*
       * No newline; see if there is more data to be read...
       */
 
-      if ((bytes = recv(http->fd, bufend, HTTP_MAX_BUFFER - http->used, 0)) < 0)
+#ifdef HAVE_LIBSSL
+      if (http->tls)
+        bytes = SSL_read((SSL *)(http->tls), bufend,
+	                 HTTP_MAX_BUFFER - http->used);
+      else
+#endif /* HAVE_LIBSSL */
+      bytes = recv(http->fd, bufend, HTTP_MAX_BUFFER - http->used, 0);
+
+      if (bytes < 0)
       {
        /*
 	* Nope, can't get a line this time...
 	*/
 
-#if defined(WIN32) || defined(__EMX__)
+#ifdef WIN32
         if (WSAGetLastError() != http->error)
 	{
 	  http->error = WSAGetLastError();
@@ -935,7 +1462,7 @@ httpGets(char   *line,			/* I - Line to read into */
 	}
 
         DEBUG_printf(("httpGets(): recv() error %d!\n", errno));
-#endif /* WIN32 || __EMX__ */
+#endif /* WIN32 */
 
         return (NULL);
       }
@@ -955,7 +1482,7 @@ httpGets(char   *line,			/* I - Line to read into */
       bufend     += bytes;
     }
   }
-  while (bufptr >= bufend);
+  while (bufptr >= bufend && http->used < HTTP_MAX_BUFFER);
 
   http->activity = time(NULL);
 
@@ -966,6 +1493,7 @@ httpGets(char   *line,			/* I - Line to read into */
   lineptr = line;
   bufptr  = http->buffer;
   bytes   = 0;
+  length --;
 
   while (bufptr < bufend && bytes < length)
   {
@@ -974,14 +1502,7 @@ httpGets(char   *line,			/* I - Line to read into */
     if (*bufptr == 0x0a)
     {
       bufptr ++;
-      *lineptr = '\0';
-
-      http->used -= bytes;
-      if (http->used > 0)
-	memcpy(http->buffer, bufptr, http->used);
-
-      DEBUG_printf(("httpGets(): Returning \"%s\"\n", line));
-      return (line);
+      break;
     }
     else if (*bufptr == 0x0d)
       bufptr ++;
@@ -989,6 +1510,18 @@ httpGets(char   *line,			/* I - Line to read into */
       *lineptr++ = *bufptr++;
   }
 
+  if (bytes > 0)
+  {
+    *lineptr = '\0';
+
+    http->used -= bytes;
+    if (http->used > 0)
+      memcpy(http->buffer, bufptr, http->used);
+
+    DEBUG_printf(("httpGets(): Returning \"%s\"\n", line));
+    return (line);
+  }
+
   DEBUG_puts("httpGets(): No new line available!");
 
   return (NULL);
@@ -1019,8 +1552,17 @@ httpPrintf(http_t     *http,		/* I - HTTP data */
   DEBUG_printf(("httpPrintf: %s", buf));
 
   for (tbytes = 0, bufptr = buf; tbytes < bytes; tbytes += nbytes, bufptr += nbytes)
-    if ((nbytes = send(http->fd, bufptr, bytes - tbytes, 0)) < 0)
+  {
+#ifdef HAVE_LIBSSL
+    if (http->tls)
+      nbytes = SSL_write((SSL *)(http->tls), bufptr, bytes - tbytes);
+    else
+#endif /* HAVE_LIBSSL */
+    nbytes = send(http->fd, bufptr, bytes - tbytes, 0);
+
+    if (nbytes < 0)
       return (-1);
+  }
 
   return (bytes);
 }
@@ -1035,6 +1577,10 @@ httpStatus(http_status_t status)	/* I - HTTP status code */
 {
   switch (status)
   {
+    case HTTP_CONTINUE :
+        return ("Continue");
+    case HTTP_SWITCHING_PROTOCOLS :
+        return ("Switching Protocols");
     case HTTP_OK :
         return ("OK");
     case HTTP_CREATED :
@@ -1057,6 +1603,8 @@ httpStatus(http_status_t status)	/* I - HTTP status code */
         return ("Request Entity Too Large");
     case HTTP_URI_TOO_LONG :
         return ("URI Too Long");
+    case HTTP_UPGRADE_REQUIRED :
+        return ("Upgrade Required");
     case HTTP_NOT_IMPLEMENTED :
         return ("Not Implemented");
     case HTTP_NOT_SUPPORTED :
@@ -1079,9 +1627,9 @@ httpGetDateString(time_t t)		/* I - UNIX time */
 
 
   tdate = gmtime(&t);
-  sprintf(datetime, "%s, %02d %s %d %02d:%02d:%02d GMT",
-          days[tdate->tm_wday], tdate->tm_mday, months[tdate->tm_mon],
-	  tdate->tm_year + 1900, tdate->tm_hour, tdate->tm_min, tdate->tm_sec);
+  snprintf(datetime, sizeof(datetime), "%s, %02d %s %d %02d:%02d:%02d GMT",
+           days[tdate->tm_wday], tdate->tm_mday, months[tdate->tm_mon],
+	   tdate->tm_year + 1900, tdate->tm_hour, tdate->tm_min, tdate->tm_sec);
 
   return (datetime);
 }
@@ -1135,6 +1683,10 @@ httpUpdate(http_t *http)		/* I - HTTP data */
   http_field_t	field;			/* Field index */
   int		major, minor;		/* HTTP version numbers */
   http_status_t	status;			/* Authorization status */
+#ifdef HAVE_LIBSSL
+  SSL_CTX	*context;		/* Context for encryption */
+  SSL		*conn;			/* Connection for encryption */
+#endif /* HAVE_LIBSSL */
 
 
   DEBUG_printf(("httpUpdate(%08x)\n", http));
@@ -1168,6 +1720,40 @@ httpUpdate(http_t *http)		/* I - HTTP data */
       if (http->status == HTTP_CONTINUE)
         return (http->status);
 
+#ifdef HAVE_LIBSSL
+      if (http->status == HTTP_SWITCHING_PROTOCOLS && !http->tls)
+      {
+	context = SSL_CTX_new(SSLv23_method());
+	conn    = SSL_new(context);
+
+	SSL_set_fd(conn, http->fd);
+	if (SSL_connect(conn) != 1)
+	{
+	  SSL_CTX_free(context);
+	  SSL_free(conn);
+
+#ifdef WIN32
+	  http->error  = WSAGetLastError();
+#else
+	  http->error  = errno;
+#endif /* WIN32 */
+	  http->status = HTTP_ERROR;
+
+#ifdef WIN32
+	  closesocket(http->fd);
+#else
+	  close(http->fd);
+#endif
+
+	  return (HTTP_ERROR);
+	}
+
+	http->tls = conn;
+
+        return (HTTP_CONTINUE);
+      }
+#endif /* HAVE_LIBSSL */
+
       httpGetLength(http);
 
       switch (http->state)
@@ -1192,7 +1778,7 @@ httpUpdate(http_t *http)		/* I - HTTP data */
       * Got the beginning of a response...
       */
 
-      if (sscanf(line, "HTTP/%d.%d%d", &major, &minor, &status) != 3)
+      if (sscanf(line, "HTTP/%d.%d%d", &major, &minor, (int *)&status) != 3)
         return (HTTP_ERROR);
 
       http->version = (http_version_t)(major * 100 + minor);
@@ -1468,12 +2054,15 @@ http_send(http_t       *http,	/* I - HTTP data */
   * Encode the URI as needed...
   */
 
-  for (ptr = buf; *uri != '\0'; uri ++)
+  for (ptr = buf; *uri != '\0' && ptr < (buf + sizeof(buf) - 1); uri ++)
     if (*uri <= ' ' || *uri >= 127)
     {
-      *ptr ++ = '%';
-      *ptr ++ = hex[(*uri >> 4) & 15];
-      *ptr ++ = hex[*uri & 15];
+      if (ptr < (buf + sizeof(buf) - 1))
+        *ptr ++ = '%';
+      if (ptr < (buf + sizeof(buf) - 1))
+        *ptr ++ = hex[(*uri >> 4) & 15];
+      if (ptr < (buf + sizeof(buf) - 1))
+        *ptr ++ = hex[*uri & 15];
     }
     else
       *ptr ++ = *uri;
@@ -1497,6 +2086,14 @@ http_send(http_t       *http,	/* I - HTTP data */
 
   http->status = HTTP_CONTINUE;
 
+#ifdef HAVE_LIBSSL
+  if (http->encryption == HTTP_ENCRYPT_REQUIRED && !http->tls)
+  {
+    httpSetField(http, HTTP_FIELD_CONNECTION, "Upgrade");
+    httpSetField(http, HTTP_FIELD_UPGRADE, "TLS/1.0,SSL/2.0,SSL/3.0");
+  }
+#endif /* HAVE_LIBSSL */
+
   if (httpPrintf(http, "%s %s HTTP/1.1\r\n", codes[request], buf) < 1)
   {
     http->status = HTTP_ERROR;
@@ -1527,6 +2124,98 @@ http_send(http_t       *http,	/* I - HTTP data */
 }
 
 
+#ifdef HAVE_LIBSSL
 /*
- * End of "$Id: http.c,v 1.57 1999/12/21 02:26:45 mike Exp $".
+ * 'http_upgrade()' - Force upgrade to TLS encryption.
+ */
+
+static int			/* O - Status of connection */
+http_upgrade(http_t *http)	/* I - HTTP data */
+{
+  int		ret;		/* Return value */
+  http_t	myhttp;		/* Local copy of HTTP data */
+
+
+  DEBUG_printf(("http_upgrade(%p)\n", http));
+
+ /*
+  * Copy the HTTP data to a local variable so we can do the OPTIONS
+  * request without interfering with the existing request data...
+  */
+
+  memcpy(&myhttp, http, sizeof(myhttp));
+
+ /*
+  * Send an OPTIONS request to the server, requiring SSL or TLS
+  * encryption on the link...
+  */
+
+  httpClearFields(&myhttp);
+  httpSetField(&myhttp, HTTP_FIELD_CONNECTION, "upgrade");
+  httpSetField(&myhttp, HTTP_FIELD_UPGRADE, "TLS/1.0, SSL/2.0, SSL/3.0");
+
+  if ((ret = httpOptions(&myhttp, "*")) == 0)
+  {
+   /*
+    * Wait for the secure connection...
+    */
+
+    while (httpUpdate(&myhttp) == HTTP_CONTINUE);
+  }
+
+  httpFlush(&myhttp);
+
+ /*
+  * Copy the HTTP data back over, if any...
+  */
+
+  http->fd         = myhttp.fd;
+  http->error      = myhttp.error;
+  http->activity   = myhttp.activity;
+  http->status     = myhttp.status;
+  http->version    = myhttp.version;
+  http->keep_alive = myhttp.keep_alive;
+  http->used       = myhttp.used;
+
+  if (http->used)
+    memcpy(http->buffer, myhttp.buffer, http->used);
+
+  http->auth_type   = myhttp.auth_type;
+  http->nonce_count = myhttp.nonce_count;
+
+  memcpy(http->nonce, myhttp.nonce, sizeof(http->nonce));
+
+  http->tls        = myhttp.tls;
+  http->encryption = myhttp.encryption;
+
+ /*
+  * See if we actually went secure...
+  */
+
+  if (!http->tls)
+  {
+   /*
+    * Server does not support HTTP upgrade...
+    */
+
+    DEBUG_puts("Server does not support HTTP upgrade!");
+
+#ifdef WIN32
+    closesocket(http->fd);
+#else
+    close(http->fd);
+#endif
+
+    http->fd = -1;
+
+    return (-1);
+  }
+  else
+    return (ret);
+}
+#endif /* HAVE_LIBSSL */
+
+
+/*
+ * End of "$Id: http.c,v 1.94 2002/02/12 19:05:38 mike Exp $".
  */
