@@ -1,9 +1,9 @@
 /*
- * "$Id: ppd.c,v 1.38 1999/12/21 02:26:46 mike Exp $"
+ * "$Id: ppd.c,v 1.59 2002/02/12 19:05:39 mike Exp $"
  *
  *   PPD file routines for the Common UNIX Printing System (CUPS).
  *
- *   Copyright 1997-1999 by Easy Software Products, all rights reserved.
+ *   Copyright 1997-2002 by Easy Software Products, all rights reserved.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -108,6 +108,7 @@ ppdClose(ppd_file_t *ppd)	/* I - PPD file record */
   ppd_emul_t	*emul;		/* Current emulation */
   ppd_group_t	*group;		/* Current group */
   char		**font;		/* Current font */
+  char		**filter;	/* Current filter */
 
 
  /*
@@ -121,6 +122,10 @@ ppdClose(ppd_file_t *ppd)	/* I - PPD file record */
   * Free all strings at the top level...
   */
 
+  safe_free(ppd->patches);
+  safe_free(ppd->jcl_begin);
+  safe_free(ppd->jcl_ps);
+  safe_free(ppd->jcl_end);
   safe_free(ppd->lang_encoding);
   safe_free(ppd->lang_version);
   safe_free(ppd->modelname);
@@ -171,6 +176,18 @@ ppdClose(ppd_file_t *ppd)	/* I - PPD file record */
   if (ppd->num_consts > 0)
     safe_free(ppd->consts);
 
+ /*
+  * Free any filters...
+  */
+
+  if (ppd->num_filters > 0)
+  {
+    for (i = ppd->num_filters, filter = ppd->filters; i > 0; i --, filter ++)
+      safe_free(*filter);
+
+    safe_free(ppd->filters);
+  }
+
  /*
   * Free any fonts...
   */
@@ -267,12 +284,16 @@ ppd_get_group(ppd_file_t *ppd,	/* I - PPD file */
   ppd_group_t	*group;		/* Group */
 
 
+  DEBUG_printf(("ppd_get_group(%p, \"%s\")\n", ppd, name));
+
   for (i = ppd->num_groups, group = ppd->groups; i > 0; i --, group ++)
     if (strcmp(group->text, name) == 0)
       break;
 
   if (i == 0)
   {
+    DEBUG_printf(("Adding group %s...\n", name));
+
     if (ppd->num_groups == 0)
       group = malloc(sizeof(ppd_group_t));
     else
@@ -411,12 +432,17 @@ ppdOpen(FILE *fp)		/* I - File to read from */
   ppd_const_t	*constraint;	/* Current constraint */
   ppd_size_t	*size;		/* Current page size */
   int		mask;		/* Line data mask */
-  char		keyword[41],	/* Keyword from file */
-		name[41],	/* Option from file */
-		text[81],	/* Human-readable text from file */
+  char		keyword[PPD_MAX_NAME],
+  				/* Keyword from file */
+		name[PPD_MAX_NAME],
+				/* Option from file */
+		text[PPD_MAX_TEXT],
+				/* Human-readable text from file */
 		*string,	/* Code/text from file */
 		*sptr,		/* Pointer into string */
-		*nameptr;	/* Pointer into name */
+		*nameptr,	/* Pointer into name */
+		*temp,		/* Temporary string pointer */
+		**tempfonts;	/* Temporary fonts pointer */
   float		order;		/* Order dependency number */
   ppd_section_t	section;	/* Order dependency section */
   ppd_profile_t	*profile;	/* Pointer to color profile */
@@ -456,7 +482,7 @@ ppdOpen(FILE *fp)		/* I - File to read from */
     return (NULL);
   }
 
-  DEBUG_printf(("ppdOpen: keyword = %s, string = %08x\n", keyword, string));
+  DEBUG_printf(("ppdOpen: keyword = %s, string = %p\n", keyword, string));
 
   safe_free(string);
 
@@ -495,7 +521,7 @@ ppdOpen(FILE *fp)		/* I - File to read from */
     if (string != NULL)
     {
       if (strlen(string) > 40)
-        printf(", string = %08x", string);
+        printf(", string = %p", string);
       else
         printf(", string = \"%s\"", string);
     }
@@ -503,6 +529,21 @@ ppdOpen(FILE *fp)		/* I - File to read from */
     puts("");
 #endif /* DEBUG */
 
+    if (strcmp(keyword, "CloseUI") != 0 &&
+        strcmp(keyword, "JCLCloseUI") != 0 &&
+        strcmp(keyword, "CloseGroup") != 0 &&
+	strcmp(keyword, "CloseSubGroup") != 0 &&
+	strncmp(keyword, "Default", 7) != 0 &&
+	string == NULL)
+    {
+     /*
+      * Need a string value!
+      */
+
+      ppdClose(ppd);
+      return (NULL);
+    }
+
     if (strcmp(keyword, "LanguageLevel") == 0)
       ppd->language_level = atoi(string);
     else if (strcmp(keyword, "LanguageEncoding") == 0)
@@ -584,6 +625,8 @@ ppdOpen(FILE *fp)		/* I - File to read from */
       else
         ppd->colorspace = PPD_CS_GRAY;
     }
+    else if (strcmp(keyword, "cupsFlipDuplex") == 0)
+      ppd->flip_duplex = strcmp(string, "True") == 0;
     else if (strcmp(keyword, "cupsManualCopies") == 0)
       ppd->manual_copies = strcmp(string, "True") == 0;
     else if (strcmp(keyword, "cupsModelNumber") == 0)
@@ -618,6 +661,13 @@ ppdOpen(FILE *fp)		/* I - File to read from */
       else
         filter = realloc(ppd->filters, sizeof(char *) * (ppd->num_filters + 1));
 
+      if (filter == NULL)
+      {
+        safe_free(filter);
+	ppdClose(ppd);
+	return (NULL);
+      }
+
       ppd->filters     = filter;
       filter           += ppd->num_filters;
       ppd->num_filters ++;
@@ -629,6 +679,31 @@ ppdOpen(FILE *fp)		/* I - File to read from */
       *filter = string;
       string  = NULL;
     }
+    else if (strcmp(keyword, "Throughput") == 0)
+      ppd->throughput = atoi(string);
+    else if (strcmp(keyword, "Font") == 0)
+    {
+     /*
+      * Add this font to the list of available fonts...
+      */
+
+      if (ppd->num_fonts == 0)
+        tempfonts = (char **)malloc(sizeof(char *));
+      else
+        tempfonts = (char **)realloc(ppd->fonts,
+	                             sizeof(char *) * (ppd->num_fonts + 1));
+
+      if (tempfonts == NULL)
+      {
+        safe_free(string);
+        ppdClose(ppd);
+	return (NULL);
+      }
+      
+      ppd->fonts                 = tempfonts;
+      ppd->fonts[ppd->num_fonts] = strdup(name);
+      ppd->num_fonts ++;
+    }
     else if (strcmp(keyword, "VariablePaperSize") == 0 &&
              strcmp(string, "True") == 0 &&
 	     !ppd->variable_sizes)
@@ -645,20 +720,26 @@ ppdOpen(FILE *fp)		/* I - File to read from */
       * Add a "Custom" page size option...
       */
 
-      if ((group = ppd_get_group(ppd,
-                                 cupsLangString(language,
-                                                CUPS_MSG_GENERAL))) == NULL)
+      if ((option = ppdFindOption(ppd, "PageSize")) == NULL)
       {
-        ppdClose(ppd);
-	safe_free(string);
-	return (NULL);
-      }
+        ppd_group_t	*temp;
 
-      if ((option = ppd_get_option(group, "PageSize")) == NULL)
-      {
-        ppdClose(ppd);
-	safe_free(string);
-	return (NULL);
+
+	if ((temp = ppd_get_group(ppd,
+                                  cupsLangString(language,
+                                                 CUPS_MSG_GENERAL))) == NULL)
+	{
+          ppdClose(ppd);
+	  safe_free(string);
+	  return (NULL);
+	}
+
+	if ((option = ppd_get_option(temp, "PageSize")) == NULL)
+	{
+          ppdClose(ppd);
+	  safe_free(string);
+	  return (NULL);
+	}
       }
 
       if ((choice = ppd_add_choice(option, "Custom")) == NULL)
@@ -670,7 +751,6 @@ ppdOpen(FILE *fp)		/* I - File to read from */
 
       strncpy(choice->text, cupsLangString(language, CUPS_MSG_VARIABLE),
               sizeof(choice->text) - 1);
-      group  = NULL;
       option = NULL;
     }
     else if (strcmp(keyword, "MaxMediaWidth") == 0)
@@ -707,24 +787,33 @@ ppdOpen(FILE *fp)		/* I - File to read from */
 	* Add a "Custom" page size option...
 	*/
 
-	if ((group = ppd_get_group(ppd,
-                                   cupsLangString(language,
-                                                  CUPS_MSG_GENERAL))) == NULL)
+        if ((option = ppdFindOption(ppd, "PageSize")) == NULL)
 	{
-          ppdClose(ppd);
-	  safe_free(string);
-	  return (NULL);
-	}
+	  ppd_group_t	*temp;
 
-	if ((option = ppd_get_option(group, "PageSize")) == NULL)
-	{
-          ppdClose(ppd);
-	  safe_free(string);
-	  return (NULL);
-	}
+
+	  if ((temp = ppd_get_group(ppd,
+                                    cupsLangString(language,
+                                                   CUPS_MSG_GENERAL))) == NULL)
+	  {
+	    DEBUG_puts("Unable to get general group!");
+            ppdClose(ppd);
+	    safe_free(string);
+	    return (NULL);
+	  }
+
+	  if ((option = ppd_get_option(temp, "PageSize")) == NULL)
+	  {
+	    DEBUG_puts("Unable to get PageSize option!");
+            ppdClose(ppd);
+	    safe_free(string);
+	    return (NULL);
+	  }
+        }
 
 	if ((choice = ppd_add_choice(option, "Custom")) == NULL)
 	{
+	  DEBUG_puts("Unable to add Custom choice!");
           ppdClose(ppd);
 	  safe_free(string);
 	  return (NULL);
@@ -732,12 +821,12 @@ ppdOpen(FILE *fp)		/* I - File to read from */
 
 	strncpy(choice->text, cupsLangString(language, CUPS_MSG_VARIABLE),
         	sizeof(choice->text) - 1);
-	group  = NULL;
 	option = NULL;
       }
 
       if ((option = ppdFindOption(ppd, "PageSize")) == NULL)
       {
+	DEBUG_puts("Unable to find PageSize option!");
 	ppdClose(ppd);
 	safe_free(string);
 	return (NULL);
@@ -745,6 +834,7 @@ ppdOpen(FILE *fp)		/* I - File to read from */
 
       if ((choice = ppdFindChoice(option, "Custom")) == NULL)
       {
+	DEBUG_puts("Unable to find Custom choice!");
         ppdClose(ppd);
 	safe_free(string);
 	return (NULL);
@@ -776,8 +866,11 @@ ppdOpen(FILE *fp)		/* I - File to read from */
 
       for (i = 0, sptr = string; i < count; i ++)
       {
-        for (nameptr = ppd->emulations[i].name; *sptr != '\0' && *sptr != ' ';)
-	  *nameptr ++ = *sptr ++;
+        for (nameptr = ppd->emulations[i].name;
+	     *sptr != '\0' && *sptr != ' ';
+	     sptr ++)
+	  if (nameptr < (ppd->emulations[i].name + sizeof(ppd->emulations[i].name) - 1))
+	    *nameptr++ = *sptr;
 
 	*nameptr = '\0';
 
@@ -816,8 +909,16 @@ ppdOpen(FILE *fp)		/* I - File to read from */
       }
       else
       {
-        ppd->patches = realloc(ppd->patches, strlen(ppd->patches) +
-	                                     strlen(string) + 1);
+        temp = realloc(ppd->patches, strlen(ppd->patches) +
+	                             strlen(string) + 1);
+        if (temp == NULL)
+	{
+	  safe_free(string);
+	  ppdClose(ppd);
+	  return (NULL);
+	}
+
+        ppd->patches = temp;
 
         strcpy(ppd->patches + strlen(ppd->patches), string);
       }
@@ -829,13 +930,13 @@ ppdOpen(FILE *fp)		/* I - File to read from */
       */
 
       if (name[0] == '*')
-        strcpy(name, name + 1);
+        strcpy(name, name + 1); /* Eliminate leading asterisk */
 
-      if (string == NULL)
-      {
-        ppdClose(ppd);
-	return (NULL);
-      }
+      for (i = strlen(name) - 1; i > 0 && isspace(name[i]); i --)
+        name[i] = '\0'; /* Eliminate trailing spaces */
+
+      DEBUG_printf(("OpenUI of %s in group %s...\n", name,
+                    group ? group->text : "(null)"));
 
       if (subgroup != NULL)
         option = ppd_get_option(subgroup, name);
@@ -864,6 +965,7 @@ ppdOpen(FILE *fp)		/* I - File to read from */
 	  return (NULL);
 	}
 
+        DEBUG_printf(("Adding to group %s...\n", group->text));
         option = ppd_get_option(group, name);
 	group  = NULL;
       }
@@ -988,43 +1090,6 @@ ppdOpen(FILE *fp)		/* I - File to read from */
     }
     else if (strcmp(keyword, "CloseGroup") == 0)
       group = NULL;
-    else if (strcmp(keyword, "OpenSubGroup") == 0)
-    {
-     /*
-      * Open a new sub-group...
-      */
-
-      if (group == NULL || subgroup != NULL)
-      {
-        ppdClose(ppd);
-	safe_free(string);
-	return (NULL);
-      }
-
-      if (group->num_subgroups == 0)
-	subgroup = malloc(sizeof(ppd_group_t));
-      else
-	subgroup = realloc(group->subgroups,
-	                   (group->num_subgroups + 1) * sizeof(ppd_group_t));
-
-      if (subgroup == NULL)
-      {
-	ppdClose(ppd);
-	safe_free(string);
-	return (NULL);
-      }
-
-      group->subgroups = subgroup;
-      subgroup += group->num_subgroups;
-      group->num_subgroups ++;
-
-      memset(subgroup, 0, sizeof(ppd_group_t));
-      ppd_decode(string);
-      ppd_fix(string);
-      strncpy(subgroup->text, string, sizeof(subgroup->text) - 1);
-    }
-    else if (strcmp(keyword, "CloseSubGroup") == 0)
-      subgroup = NULL;
     else if (strcmp(keyword, "OrderDependency") == 0 ||
              strcmp(keyword, "NonUIOrderDependency") == 0)
     {
@@ -1053,24 +1118,25 @@ ppdOpen(FILE *fp)		/* I - File to read from */
 
       if (option == NULL)
       {
+        ppd_group_t	*temp;
+
+
        /*
         * Only valid for Non-UI options...
 	*/
 
-        for (i = ppd->num_groups, group = ppd->groups; i > 0; i --, group ++)
-          if (group->text[0] == '\0')
+        for (i = ppd->num_groups, temp = ppd->groups; i > 0; i --, temp ++)
+          if (temp->text[0] == '\0')
 	    break;
 
         if (i > 0)
-          for (i = 0; i < group->num_options; i ++)
-	    if (strcmp(keyword, group->options[i].keyword) == 0)
+          for (i = 0; i < temp->num_options; i ++)
+	    if (strcmp(keyword, temp->options[i].keyword) == 0)
 	    {
-	      group->options[i].section = section;
-	      group->options[i].order   = order;
+	      temp->options[i].section = section;
+	      temp->options[i].order   = order;
 	      break;
 	    }
-
-        group = NULL;
       }
       else
       {
@@ -1088,24 +1154,25 @@ ppdOpen(FILE *fp)		/* I - File to read from */
 
       if (option == NULL)
       {
+        ppd_group_t	*temp;
+
+
        /*
         * Only valid for Non-UI options...
 	*/
 
-        for (i = ppd->num_groups, group = ppd->groups; i > 0; i --, group ++)
-          if (group->text[0] == '\0')
+        for (i = ppd->num_groups, temp = ppd->groups; i > 0; i --, temp ++)
+          if (temp->text[0] == '\0')
 	    break;
 
         if (i > 0)
-          for (i = 0; i < group->num_options; i ++)
-	    if (strcmp(keyword, group->options[i].keyword) == 0)
+          for (i = 0; i < temp->num_options; i ++)
+	    if (strcmp(keyword, temp->options[i].keyword) == 0)
 	    {
-	      strncpy(group->options[i].defchoice, string,
-                      sizeof(group->options[i].defchoice) - 1);
+	      strncpy(temp->options[i].defchoice, string,
+                      sizeof(temp->options[i].defchoice) - 1);
 	      break;
 	    }
-
-        group = NULL;
       }
       else
         strncpy(option->defchoice, string, sizeof(option->defchoice) - 1);
@@ -1141,6 +1208,11 @@ ppdOpen(FILE *fp)		/* I - File to read from */
 	    break;
 
 	case 2 : /* Two options... */
+	   /*
+	    * The following strcpy's are safe, as optionN and
+	    * choiceN are all the same size (size defined by PPD spec...)
+	    */
+
 	    if (constraint->option1[0] == '*')
 	      strcpy(constraint->option1, constraint->option1 + 1);
 
@@ -1154,6 +1226,11 @@ ppdOpen(FILE *fp)		/* I - File to read from */
 	    break;
 	    
 	case 3 : /* Two options, one choice... */
+	   /*
+	    * The following strcpy's are safe, as optionN and
+	    * choiceN are all the same size (size defined by PPD spec...)
+	    */
+
 	    if (constraint->option1[0] == '*')
 	      strcpy(constraint->option1, constraint->option1 + 1);
 
@@ -1183,26 +1260,55 @@ ppdOpen(FILE *fp)		/* I - File to read from */
     }
     else if (strcmp(keyword, "PaperDimension") == 0)
     {
-      if ((size = ppdPageSize(ppd, name)) != NULL)
-        sscanf(string, "%f%f", &(size->width), &(size->length));
+      if ((size = ppdPageSize(ppd, name)) == NULL)
+	size = ppd_add_size(ppd, name);
+
+      if (size == NULL)
+      {
+       /*
+        * Unable to add or find size!
+	*/
+
+        ppdClose(ppd);
+	safe_free(string);
+	return (NULL);
+      }
+
+      sscanf(string, "%f%f", &(size->width), &(size->length));
     }
     else if (strcmp(keyword, "ImageableArea") == 0)
     {
-      if ((size = ppdPageSize(ppd, name)) != NULL)
-	sscanf(string, "%f%f%f%f", &(size->left), &(size->bottom),
-	       &(size->right), &(size->top));
+      if ((size = ppdPageSize(ppd, name)) == NULL)
+	size = ppd_add_size(ppd, name);
+
+      if (size == NULL)
+      {
+       /*
+        * Unable to add or find size!
+	*/
+
+        ppdClose(ppd);
+	safe_free(string);
+	return (NULL);
+      }
+
+      sscanf(string, "%f%f%f%f", &(size->left), &(size->bottom),
+	     &(size->right), &(size->top));
     }
     else if (option != NULL &&
              (mask & (PPD_KEYWORD | PPD_OPTION | PPD_STRING)) ==
 	         (PPD_KEYWORD | PPD_OPTION | PPD_STRING))
     {
+      DEBUG_printf(("group = %p, subgroup = %p\n", group, subgroup));
+
       if (strcmp(keyword, "PageSize") == 0)
       {
        /*
         * Add a page size...
 	*/
 
-	ppd_add_size(ppd, name);
+        if (ppdPageSize(ppd, name) == NULL)
+	  ppd_add_size(ppd, name);
       }
 
      /*
@@ -1235,7 +1341,7 @@ ppdOpen(FILE *fp)		/* I - File to read from */
 
 #ifdef DEBUG
   if (!feof(fp))
-    printf("Premature EOF at %d...\n", ftell(fp));
+    printf("Premature EOF at %lu...\n", (unsigned long)ftell(fp));
 #endif /* DEBUG */
 
  /*
@@ -1366,7 +1472,7 @@ ppdOpenFile(const char *filename) /* I - File to read from */
  * 'compare_strings()' - Compare two strings.
  */
 
-int				/* O - Result of comparison */
+static int			/* O - Result of comparison */
 compare_strings(char *s,	/* I - First string */
                 char *t)	/* I - Second string */
 {
@@ -1520,6 +1626,7 @@ ppd_read(FILE *fp,		/* I - File to read from */
 	 char **string)		/* O - Code/string data */
 {
   int		ch,		/* Character from file */
+		colon,		/* Colon seen? */
 		endquote,	/* Waiting for an end quote */
 		mask;		/* Mask to be returned */
   char		*keyptr,	/* Keyword pointer */
@@ -1527,7 +1634,7 @@ ppd_read(FILE *fp,		/* I - File to read from */
 		*textptr,	/* Text pointer */
 		*strptr,	/* Pointer into string */
 		*lineptr,	/* Current position in line buffer */
-		line[262144];	/* Line buffer (256k) */
+		line[65536];	/* Line buffer (64k) */
 
 
  /*
@@ -1542,6 +1649,8 @@ ppd_read(FILE *fp,		/* I - File to read from */
   * Now loop until we have a valid line...
   */
 
+  *string = NULL;
+
   do
   {
    /*
@@ -1550,6 +1659,7 @@ ppd_read(FILE *fp,		/* I - File to read from */
 
     lineptr  = line;
     endquote = 0;
+    colon    = 0;
 
     while ((ch = getc(fp)) != EOF &&
            (lineptr - line) < (sizeof(line) - 1))
@@ -1575,10 +1685,12 @@ ppd_read(FILE *fp,		/* I - File to read from */
 	    ungetc(ch, fp);
 	}
 
-	*lineptr++ = '\n';
+	ch = '\n';
 
 	if (!endquote)			/* Continue for multi-line text */
           break;
+
+	*lineptr++ = '\n';
       }
       else
       {
@@ -1588,16 +1700,83 @@ ppd_read(FILE *fp,		/* I - File to read from */
 
 	*lineptr++ = ch;
 
-	if (ch == '\"')
-          endquote = !endquote;
+	if (ch == ':')
+	  colon = 1;
+
+	if (ch == '\"' && colon)
+        {
+	  endquote = !endquote;
+
+          if (!endquote)
+	  {
+	   /*
+	    * End of quoted string; ignore trailing characters...
+	    */
+
+	    while ((ch = getc(fp)) != EOF)
+	      if (ch == '\n')
+		break;
+	      else if (ch == '\r')
+	      {
+		ch = getc(fp);
+		if (ch != '\n')
+	          ungetc(ch, fp);
+		break;
+	      }
+
+            break;
+	  }
+	}
       }
     }
 
+    if (endquote)
+    {
+     /*
+      * Didn't finish this quoted string...
+      */
+
+      while ((ch = getc(fp)) != EOF)
+        if (ch == '\"')
+	  break;
+    }
+
+    if (ch != '\n')
+    {
+     /*
+      * Didn't finish this line...
+      */
+
+      while ((ch = getc(fp)) != EOF)
+	if (ch == '\r' || ch == '\n')
+	{
+	 /*
+	  * Line feed or carriage return...
+	  */
+
+	  if (ch == '\r')
+	  {
+	   /*
+            * Check for a trailing line feed...
+	    */
+
+	    if ((ch = getc(fp)) == EOF)
+	      break;
+	    if (ch != 0x0a)
+	      ungetc(ch, fp);
+	  }
+
+	  break;
+	}
+    }
+
     if (lineptr > line && lineptr[-1] == '\n')
       lineptr --;
 
     *lineptr = '\0';
 
+/*    DEBUG_printf(("LINE = \"%s\"\n", line));*/
+
     if (ch == EOF && lineptr == line)
       return (0);
 
@@ -1616,7 +1795,8 @@ ppd_read(FILE *fp,		/* I - File to read from */
     if (line[0] != '*')			/* All lines start with an asterisk */
       continue;
 
-    if (strncmp(line, "*%", 2) == 0 ||	/* Comment line */
+    if (strcmp(line, "*") == 0 ||	/* (Bad) comment line */
+        strncmp(line, "*%", 2) == 0 ||	/* Comment line */
         strncmp(line, "*?", 2) == 0 ||	/* Query line */
         strcmp(line, "*End") == 0)	/* End of multi-line string */
       continue;
@@ -1628,19 +1808,25 @@ ppd_read(FILE *fp,		/* I - File to read from */
     keyptr = keyword;
 
     while (*lineptr != '\0' && *lineptr != ':' && !isspace(*lineptr) &&
-	   (keyptr - keyword) < 40)
+	   (keyptr - keyword) < (PPD_MAX_NAME - 1))
       *keyptr++ = *lineptr++;
 
     *keyptr = '\0';
+
+    if (strcmp(keyword, "End") == 0)
+      continue;
+
     mask |= PPD_KEYWORD;
 
-    if (*lineptr == ' ' || *lineptr == '\t')
+/*    DEBUG_printf(("keyword = \"%s\", lineptr = \"%s\"\n", keyword, lineptr));*/
+
+    if (isspace(*lineptr))
     {
      /*
       * Get an option name...
       */
 
-      while (*lineptr == ' ' || *lineptr == '\t')
+      while (isspace(*lineptr))
         lineptr ++;
 
       optptr = option;
@@ -1652,6 +1838,8 @@ ppd_read(FILE *fp,		/* I - File to read from */
       *optptr = '\0';
       mask |= PPD_OPTION;
 
+/*      DEBUG_printf(("option = \"%s\", lineptr = \"%s\"\n", option, lineptr));*/
+
       if (*lineptr == '/')
       {
        /*
@@ -1663,7 +1851,7 @@ ppd_read(FILE *fp,		/* I - File to read from */
 	textptr = text;
 
 	while (*lineptr != '\0' && *lineptr != '\n' && *lineptr != ':' &&
-               (textptr - text) < 80)
+               (textptr - text) < (PPD_MAX_TEXT - 1))
 	  *textptr++ = *lineptr++;
 
 	*textptr = '\0';
@@ -1671,6 +1859,8 @@ ppd_read(FILE *fp,		/* I - File to read from */
 
 	mask |= PPD_TEXT;
       }
+
+/*      DEBUG_printf(("text = \"%s\", lineptr = \"%s\"\n", text, lineptr));*/
     }
 
     if (*lineptr == ':')
@@ -1696,6 +1886,8 @@ ppd_read(FILE *fp,		/* I - File to read from */
 
       *strptr = '\0';
 
+/*      DEBUG_printf(("string = \"%s\", lineptr = \"%s\"\n", *string, lineptr));*/
+
       mask |= PPD_STRING;
     }
   }
@@ -1810,5 +2002,5 @@ ppd_fix(char *string)		/* IO - String to fix */
 
 
 /*
- * End of "$Id: ppd.c,v 1.38 1999/12/21 02:26:46 mike Exp $".
+ * End of "$Id: ppd.c,v 1.59 2002/02/12 19:05:39 mike Exp $".
  */
