@@ -1,9 +1,9 @@
 /*
- * "$Id: hpgl-vector.c,v 1.10 1999/12/07 21:55:51 mike Exp $"
+ * "$Id: hpgl-vector.c,v 1.15 2002/01/02 17:58:58 mike Exp $"
  *
  *   HP-GL/2 vector routines for the Common UNIX Printing System (CUPS).
  *
- *   Copyright 1993-1999 by Easy Software Products.
+ *   Copyright 1993-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -377,27 +377,47 @@ PE_polyline_encoded(int     num_params,	/* I - Number of parameters */
       case '7' :
           s ++;
           base_bits = 5;
+
+#ifdef DEBUG
+          fputs("DEBUG:     7-bit\n", stderr);
+#endif /* DEBUG */
           break;
       case ':' :	/* Select pen */
           s ++;
           PenNumber = (int)decode_number(&s, base_bits, 1.0);
 	  if (PageDirty)
 	    printf("%.3f %.3f %.3f %.2f SP\n", Pens[PenNumber].rgb[0],
-		   Pens[PenNumber].rgb[PenNumber], Pens[PenNumber].rgb[2],
+		   Pens[PenNumber].rgb[1], Pens[PenNumber].rgb[2],
 		   Pens[PenNumber].width * PenScaling);
+
+#ifdef DEBUG
+          fprintf(stderr, "DEBUG:     set pen #%d\n", PenNumber);
+#endif /* DEBUG */
           break;
       case '<' :	/* Next coords are a move-to */
           draw = 0;
           s ++;
+
+#ifdef DEBUG
+          fputs("DEBUG:     moveto\n", stderr);
+#endif /* DEBUG */
           break;
       case '>' :	/* Set fractional bits */
           s ++;
           temp      = (int)decode_number(&s, base_bits, 1.0);
           frac_bits = 1.0 / (1 << temp);
+
+#ifdef DEBUG
+          fprintf(stderr, "DEBUG:     set fractional bits %d\n", temp);
+#endif /* DEBUG */
           break;
       case '=' :	/* Next coords are absolute */
           s ++;
           abscoords = 1;
+
+#ifdef DEBUG
+          fputs("DEBUG:     absolute\n", stderr);
+#endif /* DEBUG */
           break;
       default :
           if (*s >= 63)
@@ -409,6 +429,10 @@ PE_polyline_encoded(int     num_params,	/* I - Number of parameters */
             x = decode_number(&s, base_bits, frac_bits);
             y = decode_number(&s, base_bits, frac_bits);
 
+#ifdef DEBUG
+            fprintf(stderr, "DEBUG:     coords %.3f %.3f\n", x, y);
+#endif /* DEBUG */
+
             if (abscoords)
             {
 	      tx = Transform[0][0] * x + Transform[0][1] * y +
@@ -703,5 +727,5 @@ plot_points(int     num_params,	/* I - Number of parameters */
 
 
 /*
- * End of "$Id: hpgl-vector.c,v 1.10 1999/12/07 21:55:51 mike Exp $".
+ * End of "$Id: hpgl-vector.c,v 1.15 2002/01/02 17:58:58 mike Exp $".
  */
