@@ -1,9 +1,9 @@
 /*
- * "$Id: image-colorspace.c,v 1.16 1999/12/10 21:47:56 mike Exp $"
+ * "$Id: image-colorspace.c,v 1.25 2002/01/27 21:16:13 mike Exp $"
  *
  *   Colorspace conversions for the Common UNIX Printing System (CUPS).
  *
- *   Copyright 1993-1999 by Easy Software Products.
+ *   Copyright 1993-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -23,6 +23,7 @@
  *
  * Contents:
  *
+ *   ImageSetProfile()   - Set the device color profile.
  *   ImageWhiteToWhite() - Convert luminance colors to device-dependent
  *   ImageWhiteToRGB()   - Convert luminance data to RGB.
  *   ImageWhiteToBlack() - Convert luminance colors to black.
@@ -56,12 +57,30 @@
 
 
 /*
- * Globals...
+ * Define some math constants that are required...
  */
 
-extern int	ImageHaveProfile;
-extern int	ImageDensity[256];
-extern int	ImageMatrix[3][3][256];
+#ifndef M_PI
+#  define M_PI		3.14159265358979323846
+#endif /* !M_PI */
+
+#ifndef M_SQRT2
+#  define M_SQRT2	1.41421356237309504880
+#endif /* !M_SQRT2 */
+
+#ifndef M_SQRT1_2
+#  define M_SQRT1_2	0.70710678118654752440
+#endif /* !M_SQRT1_2 */
+
+
+/*
+ * Local globals...
+ */
+
+static int	ImageHaveProfile = 0;	/* Do we have a color profile? */
+static int	ImageDensity[256];	/* Ink/marker density LUT */
+static int	ImageMatrix[3][3][256];	/* Color transform matrix LUT */
+
 
 /*
  * Local functions...
@@ -78,6 +97,31 @@ static void	zrotate(float [3][3], float, float);
 static void	zshear(float [3][3], float, float);
 
 
+/*
+ * 'ImageSetProfile()' - Set the device color profile.
+ */
+
+void
+ImageSetProfile(float d,		/* I - Ink/marker density */
+                float g,		/* I - Ink/marker gamma */
+                float matrix[3][3])	/* I - Color transform matrix */
+{
+  int		i, j, k;		/* Looping vars */
+  float		m;			/* Current matrix value */
+  int		*im;			/* Pointer into ImageMatrix */
+
+  ImageHaveProfile  = 1;
+
+  for (i = 0, im = ImageMatrix[0][0]; i < 3; i ++)
+    for (j = 0; j < 3; j ++)
+      for (k = 0, m = matrix[i][j]; k < 256; k ++)
+        *im++ = (int)(k * m + 0.5);
+
+  for (k = 0, im = ImageDensity; k < 256; k ++)
+    *im++ = 255.0 * d * pow((float)k / 255.0, g) + 0.5;
+}
+
+
 /*
  * 'ImageWhiteToWhite()' - Convert luminance colors to device-dependent
  *                         luminance.
@@ -321,8 +365,7 @@ ImageRGBToCMYK(const ib_t *in,	/* I - Input pixels */
                int        count)/* I - Number of pixels */
 {
   int	c, m, y, k,		/* CMYK values */
-	km,			/* Maximum K value */
-	diff;			/* Color differences */
+	km;			/* Maximum K value */
   int	cc, cm, cy;		/* Calibrated CMY values */
 
 
@@ -335,7 +378,7 @@ ImageRGBToCMYK(const ib_t *in,	/* I - Input pixels */
       k = min(c, min(m, y));
 
       if ((km = max(c, max(m, y))) > k)
-        k = k * k / km;
+        k = k * k * k / (km * km);
 
       c -= k;
       m -= k;
@@ -385,7 +428,7 @@ ImageRGBToCMYK(const ib_t *in,	/* I - Input pixels */
       k = min(c, min(m, y));
 
       if ((km = max(c, max(m, y))) > k)
-        k = k * k / km;
+        k = k * k * k / (km * km);
 
       c -= k;
       m -= k;
@@ -879,5 +922,5 @@ zshear(float mat[3][3],	/* I - Matrix */
 
 
 /*
- * End of "$Id: image-colorspace.c,v 1.16 1999/12/10 21:47:56 mike Exp $".
+ * End of "$Id: image-colorspace.c,v 1.25 2002/01/27 21:16:13 mike Exp $".
  */
