@@ -1,9 +1,9 @@
 /*
- * "$Id: image-tiff.c,v 1.10 1999/12/03 20:01:42 mike Exp $"
+ * "$Id: image-tiff.c,v 1.20 2002/01/21 22:25:46 mike Exp $"
  *
  *   TIFF file routines for the Common UNIX Printing System (CUPS).
  *
- *   Copyright 1993-1999 by Easy Software Products.
+ *   Copyright 1993-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -35,6 +35,7 @@
 #ifdef HAVE_LIBTIFF
 #  include <tiff.h>	/* TIFF image definitions */
 #  include <tiffio.h>
+#  include <unistd.h>
 
 
 /*
@@ -53,11 +54,13 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
   TIFF		*tif;			/* TIFF file */
   uint32	width, height;		/* Size of image */
   uint16	photometric,		/* Colorspace */
+		compression,		/* Type of compression */
 		orientation,		/* Orientation */
 		resunit,		/* Units for resolution */
 		samples,		/* Number of samples/pixel */
 		bits,			/* Number of bits/pixel */
-		inkset;			/* Ink set for color separations */
+		inkset,			/* Ink set for color separations */
+		numinks;		/* Number of inks in set */
   float		xres,			/* Horizontal resolution */
 		yres;			/* Vertical resolution */
   uint16	*redcmap,		/* Red colormap information */
@@ -67,6 +70,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
 		num_colors,		/* Number of colors */
 		bpp,			/* Bytes per pixel */
 		x, y,			/* Current x & y */
+		row,			/* Current row in image */
 		xstart, ystart,		/* Starting x & y */
 		xdir, ydir,		/* X & y direction */
 		xcount, ycount,		/* X & Y counters */
@@ -93,21 +97,49 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
 
   if ((tif = TIFFFdOpen(fileno(fp), "", "r")) == NULL)
   {
+    fputs("ERROR: TIFFFdOpen() failed!\n", stderr);
     fclose(fp);
     return (-1);
   }
 
-  if (!TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width) ||
-      !TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height) ||
-      !TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric) ||
-      !TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samples) ||
-      !TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bits))
+  if (!TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width))
   {
+    fputs("ERROR: No image width tag in the file!\n", stderr);
     TIFFClose(tif);
     fclose(fp);
     return (-1);
   }
 
+  if (!TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height))
+  {
+    fputs("ERROR: No image height tag in the file!\n", stderr);
+    TIFFClose(tif);
+    fclose(fp);
+    return (-1);
+  }
+
+  if (!TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric))
+  {
+    fputs("ERROR: No photometric tag in the file!\n", stderr);
+    TIFFClose(tif);
+    fclose(fp);
+    return (-1);
+  }
+
+  if (!TIFFGetField(tif, TIFFTAG_COMPRESSION, &compression))
+  {
+    fputs("ERROR: No compression tag in the file!\n", stderr);
+    TIFFClose(tif);
+    fclose(fp);
+    return (-1);
+  }
+
+  if (!TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samples))
+    samples = 1;
+
+  if (!TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bits))
+    bits = 1;
+
  /*
   * Get the image orientation...
   */
@@ -138,6 +170,11 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
       img->xppi = 128;
       img->yppi = 128;
     }
+
+    fprintf(stderr, "DEBUG: TIFF resolution = %fx%f, units=%d\n",
+            xres, yres, resunit);
+    fprintf(stderr, "DEBUG: Stored resolution = %dx%d PPI\n",
+            img->xppi, img->yppi);
   }
 
  /*
@@ -169,6 +206,35 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
   * Set the X & Y start and direction according to the image orientation...
   */
 
+  switch (orientation)
+  {
+    case ORIENTATION_TOPRIGHT :
+        fputs("DEBUG: orientation = top-right\n", stderr);
+        break;
+    case ORIENTATION_RIGHTTOP :
+        fputs("DEBUG: orientation = right-top\n", stderr);
+        break;
+    default :
+    case ORIENTATION_TOPLEFT :
+        fputs("DEBUG: orientation = top-left\n", stderr);
+        break;
+    case ORIENTATION_LEFTTOP :
+        fputs("DEBUG: orientation = left-top\n", stderr);
+        break;
+    case ORIENTATION_BOTLEFT :
+        fputs("DEBUG: orientation = bottom-left\n", stderr);
+        break;
+    case ORIENTATION_LEFTBOT :
+        fputs("DEBUG: orientation = left-bottom\n", stderr);
+        break;
+    case ORIENTATION_BOTRIGHT :
+        fputs("DEBUG: orientation = bottom-right\n", stderr);
+        break;
+    case ORIENTATION_RIGHTBOT :
+        fputs("DEBUG: orientation = right-bottom\n", stderr);
+        break;
+  }
+
   switch (orientation)
   {
     case ORIENTATION_TOPRIGHT :
@@ -240,20 +306,23 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
   * each which must be handled separately...
   */
 
+  fprintf(stderr, "DEBUG: photometric = %d\n", photometric);
+  fprintf(stderr, "DEBUG: compression = %d\n", compression);
+
   switch (photometric)
   {
     case PHOTOMETRIC_MINISWHITE :
     case PHOTOMETRIC_MINISBLACK :
-        if (photometric == PHOTOMETRIC_MINISBLACK)
-        {
-          zero = 0;
-          one  = 255;
-        }
-        else
+        if (photometric == PHOTOMETRIC_MINISWHITE)
         {
           zero = 255;
           one  = 0;
         }
+        else
+        {
+          zero = 0;
+          one  = 255;
+        }
 
         if (orientation < ORIENTATION_LEFTTOP)
         {
@@ -261,13 +330,13 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
           * Row major order...
           */
 
-          for (y = ystart, ycount = img->ysize;
+          for (y = ystart, ycount = img->ysize, row = 0;
                ycount > 0;
-               ycount --, y += ydir)
+               ycount --, y += ydir, row ++)
           {
             if (bits == 1)
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (xcount = img->xsize, scanptr = scanline, p = in + xstart, bit = 128;
                    xcount > 0;
                    xcount --, p += pstep)
@@ -288,7 +357,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 2)
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (xcount = img->xsize, scanptr = scanline, p = in + xstart, bit = 0xc0;
                    xcount > 0;
                    xcount --, p += pstep)
@@ -309,7 +378,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 4)
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (xcount = img->xsize, scanptr = scanline, p = in + xstart, bit = 0xf0;
                    xcount > 0;
                    xcount --, p += pstep)
@@ -329,7 +398,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (xdir < 0 || zero || alpha)
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
 
               if (alpha)
 	      {
@@ -369,7 +438,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
               }
             }
             else
-              TIFFReadScanline(tif, in, y, 0);
+              TIFFReadScanline(tif, in, row, 0);
 
             if (img->colorspace == IMAGE_WHITE)
 	    {
@@ -409,13 +478,13 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
           * Column major order...
           */
 
-          for (x = xstart, xcount = img->xsize;
+          for (x = xstart, xcount = img->xsize, row = 0;
                xcount > 0;
-               xcount --, x += xdir)
+               xcount --, x += xdir, row ++)
           {
             if (bits == 1)
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (ycount = img->ysize, scanptr = scanline, p = in + ystart, bit = 128;
                    ycount > 0;
                    ycount --, p += ydir)
@@ -436,7 +505,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 2)
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (ycount = img->ysize, scanptr = scanline, p = in + ystart, bit = 0xc0;
                    ycount > 0;
                    ycount --, p += ydir)
@@ -458,7 +527,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 4)
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (ycount = img->ysize, scanptr = scanline, p = in + ystart, bit = 0xf0;
                    ycount > 0;
                    ycount --, p += ydir)
@@ -478,7 +547,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (ydir < 0 || zero || alpha)
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
 
               if (alpha)
 	      {
@@ -518,7 +587,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
 	      }
             }
             else
-              TIFFReadScanline(tif, in, x, 0);
+              TIFFReadScanline(tif, in, row, 0);
 
             if (img->colorspace == IMAGE_WHITE)
 	    {
@@ -557,6 +626,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
     case PHOTOMETRIC_PALETTE :
 	if (!TIFFGetField(tif, TIFFTAG_COLORMAP, &redcmap, &greencmap, &bluecmap))
 	{
+          fputs("ERROR: No colormap tag in the file!\n", stderr);
 	  fclose(fp);
 	  return (-1);
 	}
@@ -576,13 +646,13 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
           * Row major order...
           */
 
-          for (y = ystart, ycount = img->ysize;
+          for (y = ystart, ycount = img->ysize, row = 0;
                ycount > 0;
-               ycount --, y += ydir)
+               ycount --, y += ydir, row ++)
           {
             if (bits == 1)
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (xcount = img->xsize, scanptr = scanline,
 	               p = in + xstart * 3, bit = 128;
                    xcount > 0;
@@ -612,7 +682,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 2)
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (xcount = img->xsize, scanptr = scanline,
 	               p = in + xstart * 3, bit = 0xc0;
                    xcount > 0;
@@ -637,7 +707,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 4)
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (xcount = img->xsize, scanptr = scanline,
 	               p = in + 3 * xstart, bit = 0xf0;
                    xcount > 0;
@@ -663,7 +733,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
 
               for (xcount = img->xsize, p = in + 3 * xstart, scanptr = scanline;
                    xcount > 0;
@@ -713,13 +783,13 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
           * Column major order...
           */
 
-          for (x = xstart, xcount = img->xsize;
+          for (x = xstart, xcount = img->xsize, row = 0;
                xcount > 0;
-               xcount --, x += xdir)
+               xcount --, x += xdir, row ++)
           {
             if (bits == 1)
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (ycount = img->ysize, scanptr = scanline,
 	               p = in + 3 * ystart, bit = 128;
                    ycount > 0;
@@ -749,7 +819,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 2)
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (ycount = img->ysize, scanptr = scanline,
 	               p = in + 3 * ystart, bit = 0xc0;
                    ycount > 0;
@@ -774,7 +844,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 4)
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (ycount = img->ysize, scanptr = scanline,
 	               p = in + 3 * ystart, bit = 0xf0;
                    ycount > 0;
@@ -800,7 +870,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
 
               for (ycount = img->ysize, p = in + 3 * ystart, scanptr = scanline;
                    ycount > 0;
@@ -853,13 +923,13 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
           * Row major order...
           */
 
-          for (y = ystart, ycount = img->ysize;
+          for (y = ystart, ycount = img->ysize, row = 0;
                ycount > 0;
-               ycount --, y += ydir)
+               ycount --, y += ydir, row ++)
           {
             if (bits == 1)
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (xcount = img->xsize, scanptr = scanline, p = in + xstart * 3, bit = 0xf0;
                    xcount > 0;
                    xcount --, p += pstep)
@@ -890,7 +960,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 2)
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (xcount = img->xsize, scanptr = scanline, p = in + xstart * 3;
                    xcount > 0;
                    xcount --, p += pstep, scanptr ++)
@@ -905,7 +975,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 4)
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (xcount = img->xsize, scanptr = scanline, p = in + xstart * 3;
                    xcount > 0;
                    xcount -= 2, p += 2 * pstep, scanptr += 3)
@@ -929,7 +999,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (xdir < 0 || alpha)
             {
-              TIFFReadScanline(tif, scanline, y, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
 
               if (alpha)
 	      {
@@ -955,7 +1025,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
 	      }
             }
             else
-              TIFFReadScanline(tif, in, y, 0);
+              TIFFReadScanline(tif, in, row, 0);
 
             if ((saturation != 100 || hue != 0) && bpp > 1)
               ImageRGBAdjust(in, img->xsize, saturation, hue);
@@ -998,31 +1068,31 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
           * Column major order...
           */
 
-          for (x = xstart, xcount = img->xsize;
+          for (x = xstart, xcount = img->xsize, row = 0;
                xcount > 0;
-               xcount --, x += xdir)
+               xcount --, x += xdir, row ++)
           {
             if (bits == 1)
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (ycount = img->ysize, scanptr = scanline, p = in + ystart * 3, bit = 0xf0;
                    ycount > 0;
                    ycount --, p += pstep)
               {
         	if (*scanptr & bit & 0x88)
-                  p[0] = one;
+                  p[0] = 255;
                 else
-                  p[0] = zero;
+                  p[0] = 0;
 
         	if (*scanptr & bit & 0x44)
-                  p[1] = one;
+                  p[1] = 255;
                 else
-                  p[1] = zero;
+                  p[1] = 0;
 
         	if (*scanptr & bit & 0x22)
-                  p[2] = one;
+                  p[2] = 255;
                 else
-                  p[2] = zero;
+                  p[2] = 0;
 
         	if (bit == 0xf0)
                   bit = 0x0f;
@@ -1035,7 +1105,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 2)
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (ycount = img->ysize, scanptr = scanline, p = in + ystart * 3;
                    ycount > 0;
                    ycount --, p += pstep, scanptr ++)
@@ -1050,7 +1120,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (bits == 4)
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
               for (ycount = img->ysize, scanptr = scanline, p = in + ystart * 3;
                    ycount > 0;
                    ycount -= 2, p += 2 * pstep, scanptr += 3)
@@ -1074,7 +1144,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             }
             else if (ydir < 0 || alpha)
             {
-              TIFFReadScanline(tif, scanline, x, 0);
+              TIFFReadScanline(tif, scanline, row, 0);
 
               if (alpha)
 	      {
@@ -1100,7 +1170,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
 	      }
             }
             else
-              TIFFReadScanline(tif, in, x, 0);
+              TIFFReadScanline(tif, in, row, 0);
 
             if ((saturation != 100 || hue != 0) && bpp > 1)
               ImageRGBAdjust(in, img->ysize, saturation, hue);
@@ -1140,9 +1210,15 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
         break;
 
     case PHOTOMETRIC_SEPARATED :
-        TIFFGetField(tif, TIFFTAG_INKSET, &inkset);
+        if (!TIFFGetField(tif, TIFFTAG_INKSET, &inkset) &&
+	    !TIFFGetField(tif, TIFFTAG_NUMBEROFINKS, &numinks))
+	{
+          fputs("ERROR: No inkset or number-of-inks tag in the file!\n", stderr);
+	  fclose(fp);
+	  return (-1);
+	}
 
-	if (inkset == INKSET_CMYK)
+	if (inkset == INKSET_CMYK || numinks == 4)
 	{
           if (orientation < ORIENTATION_LEFTTOP)
           {
@@ -1150,13 +1226,13 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             * Row major order...
             */
 
-            for (y = ystart, ycount = img->ysize;
+            for (y = ystart, ycount = img->ysize, row = 0;
         	 ycount > 0;
-        	 ycount --, y += ydir)
+        	 ycount --, y += ydir, row ++)
             {
               if (bits == 1)
               {
-        	TIFFReadScanline(tif, scanline, y, 0);
+        	TIFFReadScanline(tif, scanline, row, 0);
         	for (xcount = img->xsize, scanptr = scanline, p = in + xstart * 3, bit = 0xf0;
                      xcount > 0;
                      xcount --, p += pstep)
@@ -1196,7 +1272,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
               }
               else if (bits == 2)
               {
-        	TIFFReadScanline(tif, scanline, y, 0);
+        	TIFFReadScanline(tif, scanline, row, 0);
         	for (xcount = img->xsize, scanptr = scanline, p = in + xstart * 3;
                      xcount > 0;
                      xcount --, p += pstep, scanptr ++)
@@ -1242,7 +1318,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
               }
               else if (bits == 4)
               {
-        	TIFFReadScanline(tif, scanline, y, 0);
+        	TIFFReadScanline(tif, scanline, row, 0);
         	for (xcount = img->xsize, scanptr = scanline, p = in + xstart * 3;
                      xcount > 0;
                      xcount --, p += pstep, scanptr += 2)
@@ -1286,9 +1362,14 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
                   }
         	}
               }
-              else
+              else if (img->colorspace == IMAGE_CMYK)
+	      {
+	        TIFFReadScanline(tif, scanline, row, 0);
+		ImagePutRow(img, 0, y, img->xsize, scanline);
+	      }
+	      else
               {
-        	TIFFReadScanline(tif, scanline, y, 0);
+        	TIFFReadScanline(tif, scanline, row, 0);
 
         	for (xcount = img->xsize, p = in + xstart * 3, scanptr = scanline;
                      xcount > 0;
@@ -1371,13 +1452,13 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
             * Column major order...
             */
 
-            for (x = xstart, xcount = img->xsize;
+            for (x = xstart, xcount = img->xsize, row = 0;
         	 xcount > 0;
-        	 xcount --, x += xdir)
+        	 xcount --, x += xdir, row ++)
             {
               if (bits == 1)
               {
-        	TIFFReadScanline(tif, scanline, x, 0);
+        	TIFFReadScanline(tif, scanline, row, 0);
         	for (ycount = img->ysize, scanptr = scanline, p = in + xstart * 3, bit = 0xf0;
                      ycount > 0;
                      ycount --, p += pstep)
@@ -1417,7 +1498,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
               }
               else if (bits == 2)
               {
-        	TIFFReadScanline(tif, scanline, x, 0);
+        	TIFFReadScanline(tif, scanline, row, 0);
         	for (ycount = img->ysize, scanptr = scanline, p = in + xstart * 3;
                      ycount > 0;
                      ycount --, p += pstep, scanptr ++)
@@ -1463,7 +1544,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
               }
               else if (bits == 4)
               {
-        	TIFFReadScanline(tif, scanline, x, 0);
+        	TIFFReadScanline(tif, scanline, row, 0);
         	for (ycount = img->ysize, scanptr = scanline, p = in + xstart * 3;
                      ycount > 0;
                      ycount --, p += pstep, scanptr += 2)
@@ -1507,9 +1588,14 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
                   }
         	}
               }
+              else if (img->colorspace == IMAGE_CMYK)
+	      {
+	        TIFFReadScanline(tif, scanline, row, 0);
+		ImagePutCol(img, x, 0, img->ysize, scanline);
+	      }
               else
               {
-        	TIFFReadScanline(tif, scanline, x, 0);
+        	TIFFReadScanline(tif, scanline, row, 0);
 
         	for (ycount = img->ysize, p = in + xstart * 3, scanptr = scanline;
                      ycount > 0;
@@ -1596,6 +1682,7 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
 	free(out);
 
 	TIFFClose(tif);
+	fputs("ERROR: Unknown TIFF photometric value!\n", stderr);
 	return (-1);
   }
 
@@ -1616,5 +1703,5 @@ ImageReadTIFF(image_t    *img,		/* IO - Image */
 
 
 /*
- * End of "$Id: image-tiff.c,v 1.10 1999/12/03 20:01:42 mike Exp $".
+ * End of "$Id: image-tiff.c,v 1.20 2002/01/21 22:25:46 mike Exp $".
  */
