@@ -1,9 +1,9 @@
 /*
- * "$Id: image.c,v 1.19 1999/12/01 20:25:03 mike Exp $"
+ * "$Id: image.c,v 1.30 2002/01/02 17:58:59 mike Exp $"
  *
  *   Base image support for the Common UNIX Printing System (CUPS).
  *
- *   Copyright 1993-1999 by Easy Software Products.
+ *   Copyright 1993-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -26,7 +26,6 @@
  *   ImageOpen()        - Open an image file and read it into memory.
  *   ImageClose()       - Close an image file.
  *   ImageSetMaxTiles() - Set the maximum number of tiles to cache.
- *   ImageSetProfile()  - Set the device color profile.
  *   ImageGetCol()      - Get a column of pixels from an image.
  *   ImageGetRow()      - Get a row of pixels from an image.
  *   ImagePutCol()      - Put a column of pixels to an image.
@@ -43,15 +42,7 @@
 #include <unistd.h>
 #include <ctype.h>
 #include <math.h>
-
-
-/*
- * Globals...
- */
-
-int	ImageHaveProfile = 0;	/* Do we have a color profile? */
-int	ImageDensity[256];	/* Ink/marker density LUT */
-int	ImageMatrix[3][3][256];	/* Color transform matrix LUT */
+#include <cups/cups.h>
 
 
 /*
@@ -81,22 +72,34 @@ ImageOpen(char       *filename,	/* I - Filename of image */
   int		status;		/* Status of load... */
 
 
+  fprintf(stderr, "DEBUG: ImageOpen(\"%s\", %d, %d, %d, %d, %p)\n",
+          filename ? filename : "(null)", primary, secondary,
+	  saturation, hue, lut);
+
  /*
   * Range check...
   */
 
   if (filename == NULL)
+  {
+    fputs("ERROR: Image filename == NULL!\n", stderr);
     return (NULL);
+  }
 
  /*
   * Figure out the file type...
   */
 
   if ((fp = fopen(filename, "r")) == NULL)
+  {
+    perror("ERROR: Unable to open image file");
     return (NULL);
+  }
 
   if (fread(header, 1, sizeof(header), fp) == 0)
   {
+    perror("ERROR: Unable to read image file header");
+
     fclose(fp);
     return (NULL);
   }
@@ -114,6 +117,7 @@ ImageOpen(char       *filename,	/* I - Filename of image */
 
   if (img == NULL)
   {
+    perror("ERROR: Unable to allocate memory for image file");
     fclose(fp);
     return (NULL);
   }
@@ -129,6 +133,8 @@ ImageOpen(char       *filename,	/* I - Filename of image */
   if (memcmp(header, "GIF87a", 6) == 0 ||
            memcmp(header, "GIF89a", 6) == 0)
     status = ImageReadGIF(img, fp, primary, secondary, saturation, hue, lut);
+  else if (memcmp(header, "BM", 2) == 0)
+    status = ImageReadBMP(img, fp, primary, secondary, saturation, hue, lut);
   else if (header[0] == 0x01 && header[1] == 0xda)
     status = ImageReadSGI(img, fp, primary, secondary, saturation, hue, lut);
   else if (header[0] == 0x59 && header[1] == 0xa6 &&
@@ -138,12 +144,16 @@ ImageOpen(char       *filename,	/* I - Filename of image */
     status = ImageReadPNM(img, fp, primary, secondary, saturation, hue, lut);
   else if (memcmp(header2, "PCD_IPI", 7) == 0)
     status = ImageReadPhotoCD(img, fp, primary, secondary, saturation, hue, lut);
+  else if (memcmp(header + 8, "\000\010", 2) == 0 ||
+           memcmp(header + 8, "\000\030", 2) == 0)
+    status = ImageReadPIX(img, fp, primary, secondary, saturation, hue, lut);
 #if defined(HAVE_LIBPNG) && defined(HAVE_LIBZ)
   else if (memcmp(header, "\211PNG", 4) == 0)
     status = ImageReadPNG(img, fp, primary, secondary, saturation, hue, lut);
 #endif /* HAVE_LIBPNG && HAVE_LIBZ */
 #ifdef HAVE_LIBJPEG
-  else if (memcmp(header + 6, "JFIF", 4) == 0)
+  else if (memcmp(header, "\377\330\377", 3) == 0 &&	/* Start-of-Image */
+	   header[3] >= 0xe0 && header[3] <= 0xef)	/* APPn */
     status = ImageReadJPEG(img, fp, primary, secondary, saturation, hue, lut);
 #endif /* HAVE_LIBJPEG */
 #ifdef HAVE_LIBTIFF
@@ -153,6 +163,7 @@ ImageOpen(char       *filename,	/* I - Filename of image */
 #endif /* HAVE_LIBTIFF */
   else
   {
+    fputs("ERROR: Unknown image file format!\n", stderr);
     fclose(fp);
     status = -1;
   }
@@ -197,9 +208,9 @@ ImageClose(image_t *img)	/* I - Image to close */
 
   fputs("DEBUG: Freeing memory...\n", stderr);
 
-  for (current = img->first; current != NULL; current = next)
+  for (current = img->first, next = NULL; current != NULL; current = next)
   {
-    fprintf(stderr, "DEBUG: Freeing cache (%08lx, next = %08lx)...\n",
+    fprintf(stderr, "DEBUG: Freeing cache (%p, next = %p)...\n",
             current, next);
 
     next = current->next;
@@ -212,11 +223,11 @@ ImageClose(image_t *img)	/* I - Image to close */
 
   if (img->tiles != NULL)
   {
-    fprintf(stderr, "DEBUG: Freeing tiles (%08lx)...\n", img->tiles[0]);
+    fprintf(stderr, "DEBUG: Freeing tiles (%p)...\n", img->tiles[0]);
 
     free(img->tiles[0]);
 
-    fprintf(stderr, "DEBUG: Freeing tile pointers (%08lx)...\n", img->tiles);
+    fprintf(stderr, "DEBUG: Freeing tile pointers (%p)...\n", img->tiles);
 
     free(img->tiles);
   }
@@ -290,31 +301,6 @@ ImageSetMaxTiles(image_t *img,		/* I - Image to set */
 }
 
 
-/*
- * 'ImageSetProfile()' - Set the device color profile.
- */
-
-void
-ImageSetProfile(float d,		/* I - Ink/marker density */
-                float g,		/* I - Ink/marker gamma */
-                float matrix[3][3])	/* I - Color transform matrix */
-{
-  int		i, j, k;		/* Looping vars */
-  float		m;			/* Current matrix value */
-  int		*im;			/* Pointer into ImageMatrix */
-
-  ImageHaveProfile  = 1;
-
-  for (i = 0, im = ImageMatrix[0][0]; i < 3; i ++)
-    for (j = 0; j < 3; j ++)
-      for (k = 0, m = matrix[i][j]; k < 256; k ++)
-        *im++ = (int)(k * m + 0.5);
-
-  for (k = 0, im = ImageDensity; k < 256; k ++)
-    *im++ = 255.0 * d * pow((float)k / 255.0, g) + 0.5;
-}
-
-
 /*
  * 'ImageGetCol()' - Get a column of pixels from an image.
  */
@@ -629,12 +615,12 @@ get_tile(image_t *img,	/* I - Image */
 
       img->num_ics ++;
 
-      fprintf(stderr, "DEBUG: Allocated cache tile %d (%08lx)...\n",
+      fprintf(stderr, "DEBUG: Allocated cache tile %d (%p)...\n",
               img->num_ics, ic);
     }
     else
     {
-      fprintf(stderr, "DEBUG: Flushing old cache tile (%08lx)...\n",
+      fprintf(stderr, "DEBUG: Flushing old cache tile (%p)...\n",
               img->first);
 
       flush_tile(img);
@@ -646,7 +632,7 @@ get_tile(image_t *img,	/* I - Image */
 
     if (tile->pos >= 0)
     {
-      fprintf(stderr, "DEBUG: Loading cache tile from file position %d...\n",
+      fprintf(stderr, "DEBUG: Loading cache tile from file position %ld...\n",
               tile->pos);
 
       if (ftell(img->cachefile) != tile->pos)
@@ -710,11 +696,11 @@ get_tile(image_t *img,	/* I - Image */
 static void
 flush_tile(image_t *img)	/* I - Image */
 {
+  int		fd;		/* Cache file descriptor */
   int		bpp;		/* Bytes per pixel */
   itile_t	*tile;		/* Pointer to tile */
 
 
-
   bpp  = ImageGetDepth(img);
   tile = img->first->tile;
 
@@ -726,17 +712,25 @@ flush_tile(image_t *img)	/* I - Image */
 
   if (img->cachefile == NULL)
   {
-    cupsTempFile(img->cachename, sizeof(img->cachename));
-
-    fprintf(stderr, "DEBUG: Creating swap file \"%s\"...\n", img->cachename);
-
-    if ((img->cachefile = fopen(img->cachename, "wb+")) == NULL)
+    if ((fd = cupsTempFd(img->cachename, sizeof(img->cachename))) < 0)
     {
       perror("ERROR: Unable to create image swap file");
       tile->ic    = NULL;
       tile->dirty = 0;
       return;
     }
+
+    fprintf(stderr, "DEBUG: Created swap file \"%s\"...\n", img->cachename);
+
+    if ((img->cachefile = fdopen(fd, "wb+")) == NULL)
+    {
+      perror("ERROR: Unable to create image swap file");
+      close(fd);
+      unlink(img->cachename);
+      tile->ic    = NULL;
+      tile->dirty = 0;
+      return;
+    }
   }
 
   if (tile->pos >= 0)
@@ -767,7 +761,7 @@ flush_tile(image_t *img)	/* I - Image */
   if (fwrite(tile->ic->pixels, bpp, TILE_SIZE * TILE_SIZE, img->cachefile) < 1)
     perror("ERROR: Unable to write tile to swap file");
   else
-    fprintf(stderr, "DEBUG: Wrote tile at position %d...\n", tile->pos);
+    fprintf(stderr, "DEBUG: Wrote tile at position %ld...\n", tile->pos);
 
   tile->ic    = NULL;
   tile->dirty = 0;
@@ -775,5 +769,5 @@ flush_tile(image_t *img)	/* I - Image */
 
 
 /*
- * End of "$Id: image.c,v 1.19 1999/12/01 20:25:03 mike Exp $".
+ * End of "$Id: image.c,v 1.30 2002/01/02 17:58:59 mike Exp $".
  */
