@@ -1,9 +1,9 @@
 /*
- * "$Id: pstops.c,v 1.32 1999/12/07 19:05:45 mike Exp $"
+ * "$Id: pstops.c,v 1.72 2002/01/23 22:30:42 mike Exp $"
  *
  *   PostScript filter for the Common UNIX Printing System (CUPS).
  *
- *   Copyright 1993-1999 by Easy Software Products.
+ *   Copyright 1993-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -59,7 +59,8 @@ int		Order = 0,		/* 0 = normal, 1 = reverse pages */
 		Flip = 0,		/* Flip/mirror pages */
 		NUp = 1,		/* Number of pages on each sheet (1, 2, 4) */
 		Collate = 0,		/* Collate copies? */
-		Copies = 1;		/* Number of copies */
+		Copies = 1,		/* Number of copies */
+		UseESPsp = 0;		/* Use ESPshowpage? */
 
 
 /*
@@ -88,9 +89,11 @@ main(int  argc,			/* I - Number of command-line arguments */
   const char	*val;		/* Option value */
   char		tempfile[255];	/* Temporary file name */
   FILE		*temp;		/* Temporary file */
+  int		tempfd;		/* Temporary file descriptor */
   int		number;		/* Page number */
   int		slowcollate;	/* 1 if we need to collate manually */
   int		sloworder;	/* 1 if we need to order manually */
+  int		slowduplex;	/* 1 if we need an even number of pages */
   char		line[8192];	/* Line buffer */
   float		g;		/* Gamma correction value */
   float		b;		/* Brightness factor */
@@ -98,11 +101,24 @@ main(int  argc,			/* I - Number of command-line arguments */
   int		nbytes,		/* Number of bytes read */
 		tbytes;		/* Total bytes to read for binary data */
   int		page;		/* Current page sequence number */
+  int		real_page;	/* "Real" page number in document */
   int		page_count;	/* Page count for NUp */
+  int		basepage;	/* Base page number */
   int		subpage;	/* Sub-page number */
   int		copy;		/* Current copy */
+  int		saweof;		/* Did we see a %%EOF tag? */
 
 
+ /*
+  * Make sure status messages are not buffered...
+  */
+
+  setbuf(stderr, NULL);
+
+ /*
+  * Check command-line...
+  */
+
   if (argc < 6 || argc > 7)
   {
     fputs("ERROR: pstops job-id user title copies options [file]\n", stderr);
@@ -124,7 +140,8 @@ main(int  argc,			/* I - Number of command-line arguments */
 
     if ((fp = fopen(argv[6], "rb")) == NULL)
     {
-      perror("ERROR: unable to open print file - ");
+      fprintf(stderr, "ERROR: unable to open print file \"%s\" - %s\n",
+              argv[6], strerror(errno));
       return (1);
     }
   }
@@ -138,16 +155,11 @@ main(int  argc,			/* I - Number of command-line arguments */
 
   Copies = atoi(argv[4]);
 
-  ppd = ppdOpenFile(getenv("PPD"));
-
   options     = NULL;
   num_options = cupsParseOptions(argv[5], 0, &options);
 
   ppd = SetCommonOptions(num_options, options, 1);
 
-  ppdMarkDefaults(ppd);
-  cupsMarkOptions(ppd, num_options, options);
-
   if ((val = cupsGetOption("page-ranges", num_options, options)) != NULL)
     PageRanges = val;
 
@@ -162,10 +174,10 @@ main(int  argc,			/* I - Number of command-line arguments */
     *   single-document, separate-documents-collated-copies, and
     *   single-document-new-sheet all require collated copies.
     *
-    *   separate-documents-collated-copies allows for uncollated copies.
+    *   separate-documents-uncollated-copies allows for uncollated copies.
     */
 
-    Collate = strcasecmp(val, "separate-documents-collated-copies") != 0;
+    Collate = strcasecmp(val, "separate-documents-uncollated-copies") != 0;
   }
 
   if ((val = cupsGetOption("Collate", num_options, options)) != NULL &&
@@ -189,6 +201,22 @@ main(int  argc,			/* I - Number of command-line arguments */
   * See if we have to filter the fast or slow way...
   */
 
+  if (ppd && ppd->manual_copies && Duplex && Copies > 1)
+  {
+   /*
+    * Force collated copies when printing a duplexed document to
+    * a non-PS printer that doesn't do hardware copy generation.
+    * Otherwise the copies will end up on the front/back side of
+    * each page.  Also, set the "slowduplex" option to make sure
+    * that we output an even number of pages...
+    */
+
+    Collate    = 1;
+    slowduplex = 1;
+  }
+  else
+    slowduplex = 0;
+
   if (ppdFindOption(ppd, "Collate") == NULL && Collate && Copies > 1)
     slowcollate = 1;
   else
@@ -208,11 +236,14 @@ main(int  argc,			/* I - Number of command-line arguments */
 
   if (sloworder || slowcollate)
   {
-    temp = fopen(cupsTempFile(tempfile, sizeof(tempfile)), "wb+");
+    tempfd = cupsTempFd(tempfile, sizeof(tempfile));
+    temp   = fdopen(tempfd, "wb+");
 
     if (temp == NULL)
       slowcollate = sloworder = 0;
   }
+  else
+    temp = NULL;
 
  /*
   * Write any "exit server" options that have been selected...
@@ -224,12 +255,7 @@ main(int  argc,			/* I - Number of command-line arguments */
   * Write any JCL commands that are needed to print PostScript code...
   */
 
-  if (ppd != NULL && ppd->jcl_begin && ppd->jcl_ps)
-  {
-    fputs(ppd->jcl_begin, stdout);
-    ppdEmit(ppd, stdout, PPD_ORDER_JCL);
-    fputs(ppd->jcl_ps, stdout);
-  }
+  ppdEmitJCL(ppd, stdout, atoi(argv[1]), argv[2], argv[3]);
 
  /*
   * Read the first line to see if we have DSC comments...
@@ -242,12 +268,20 @@ main(int  argc,			/* I - Number of command-line arguments */
     return (1);
   }
 
+ /*
+  * See if this is an EPS file...
+  */
+
+  UseESPsp = strstr(line, "EPS") != NULL;
+
  /*
   * Start sending the document with any commands needed...
   */
 
   puts(line);
 
+  saweof = 0;
+
   if (ppd != NULL && ppd->patches != NULL)
     puts(ppd->patches);
 
@@ -255,18 +289,59 @@ main(int  argc,			/* I - Number of command-line arguments */
   ppdEmit(ppd, stdout, PPD_ORDER_ANY);
   ppdEmit(ppd, stdout, PPD_ORDER_PROLOG);
 
-  if (NUp > 1)
-    puts("userdict begin\n"
-         "/ESPshowpage /showpage load def\n"
-         "/showpage { } def\n"
-         "end");
-
   if (g != 1.0 || b != 1.0)
     printf("{ neg 1 add dup 0 lt { pop 1 } { %.3f exp neg 1 add } "
            "ifelse %.3f mul } bind settransfer\n", g, b);
 
+ /*
+  * Figure out if we should use ESPshowpage or not...
+  */
+
+  val = cupsGetOption("page-label", num_options, options);
+
+  if (val != NULL || getenv("CLASSIFICATION") != NULL || NUp > 1)
+  {
+   /*
+    * Yes, use ESPshowpage...
+    */
+
+    UseESPsp = 1;
+  }
+
+  if (NUp == 2)
+  {
+   /*
+    * For 2-up output, rotate the labels to match the orientation
+    * of the pages...
+    */
+
+    if (Orientation & 1)
+      WriteLabelProlog(val, PageBottom, PageWidth - PageLength + PageTop,
+                       PageLength);
+    else
+      WriteLabelProlog(val, PageLeft, PageRight, PageLength);
+  }
+  else
+    WriteLabelProlog(val, PageBottom, PageTop, PageWidth);
+
+  if (UseESPsp)
+    puts("userdict begin\n"
+	 "/ESPshowpage /showpage load def\n"
+	 "/showpage { } def\n"
+	 "end");
+
   if (Copies > 1 && (!Collate || !slowcollate))
-    printf("/#copies %d def\n", Copies);
+  {
+    if (Collate)
+      printf("%%%%Requirements: numcopies(%d) collate\n", Copies);
+    else
+      printf("%%%%Requirements: numcopies(%d)\n", Copies);
+
+    if (LanguageLevel == 1)
+      printf("/#copies %d def\n", Copies);
+    else
+      printf("<</NumCopies %d>>setpagedevice\n", Copies);
+  }
 
   if (strncmp(line, "%!PS-Adobe-", 11) == 0)
   {
@@ -277,119 +352,163 @@ main(int  argc,			/* I - Number of command-line arguments */
     level = 0;
 
     while (psgets(line, sizeof(line), fp) != NULL)
+    {
+      if (strncmp(line, "%%", 2) == 0)
+        fprintf(stderr, "DEBUG: %d %s", level, line);
+
       if (strncmp(line, "%%BeginDocument:", 16) == 0 ||
           strncmp(line, "%%BeginDocument ", 16) == 0)	/* Adobe Acrobat BUG */
         level ++;
-      else if (strcmp(line, "%%EndDocument") == 0 && level > 0)
+      else if (strncmp(line, "%%EndDocument", 13) == 0 && level > 0)
         level --;
       else if (strncmp(line, "%%Page:", 7) == 0 && level == 0)
         break;
       else if (strncmp(line, "%%BeginBinary:", 14) == 0 ||
                (strncmp(line, "%%BeginData:", 12) == 0 &&
-	        strstr(line, "Binary") != NULL))
+	        strstr(line, "ASCII") == NULL && strstr(line, "Hex") == NULL))
       {
        /*
         * Copy binary data...
 	*/
 
         tbytes = atoi(strchr(line, ':') + 1);
+	fputs(line, stdout);
+
 	while (tbytes > 0)
 	{
-	  nbytes = fread(line, 1, sizeof(line), fp);
+	  if (tbytes > sizeof(line))
+	    nbytes = fread(line, 1, sizeof(line), fp);
+	  else
+	    nbytes = fread(line, 1, tbytes, fp);
+
+          if (nbytes < 1)
+	  {
+	    perror("ERROR: Early end-of-file while reading binary data");
+	    return (1);
+	  }
+
 	  fwrite(line, 1, nbytes, stdout);
 	  tbytes -= nbytes;
 	}
       }
       else
-        puts(line);
+        fputs(line, stdout);
+    }
 
    /*
     * Then read all of the pages, filtering as needed...
     */
 
-    for (page = 1;;)
+    for (page = 1, real_page = 1;;)
     {
+      if (strncmp(line, "%%", 2) == 0)
+        fprintf(stderr, "DEBUG: %d %s", level, line);
+
       if (strncmp(line, "%%BeginDocument:", 16) == 0 ||
           strncmp(line, "%%BeginDocument ", 16) == 0)	/* Adobe Acrobat BUG */
         level ++;
-      else if (strcmp(line, "%%EndDocument") == 0 && level > 0)
+      else if (strncmp(line, "%%EndDocument", 13) == 0 && level > 0)
         level --;
+      else if (strcmp(line, "\004") == 0)
+        break;
+      else if (strncmp(line, "%%EOF", 5) == 0 && level == 0)
+      {
+        fputs("DEBUG: Saw EOF!\n", stderr);
+        saweof = 1;
+	break;
+      }
       else if (strncmp(line, "%%Page:", 7) == 0 && level == 0)
       {
-        if (sscanf(line, "%*s%*s%d", &number) == 1)
+	if (!check_range(real_page))
 	{
-	  if (!check_range(number))
-	  {
-	    while (psgets(line, sizeof(line), fp) != NULL)
-	      if (strncmp(line, "%%BeginDocument:", 16) == 0 ||
-        	  strncmp(line, "%%BeginDocument ", 16) == 0)	/* Adobe Acrobat BUG */
-        	level ++;
-	      else if (strcmp(line, "%%EndDocument") == 0 && level > 0)
-        	level --;
-	      else if (strncmp(line, "%%Page:", 7) == 0 && level == 0)
-	        break;
-
-            continue;
-          }
-
-          if (!sloworder && NumPages > 0)
-	    end_nup(NumPages - 1);
-
-	  if (slowcollate || sloworder)
-	    Pages[NumPages] = ftell(temp);
-
-          if (!sloworder)
-	  {
-	    if ((NumPages & (NUp - 1)) == 0)
+	  while (psgets(line, sizeof(line), fp) != NULL)
+	    if (strncmp(line, "%%BeginDocument:", 16) == 0 ||
+        	strncmp(line, "%%BeginDocument ", 16) == 0)	/* Adobe Acrobat BUG */
+              level ++;
+	    else if (strcmp(line, "%%EndDocument") == 0 && level > 0)
+              level --;
+	    else if (strncmp(line, "%%Page:", 7) == 0 && level == 0)
 	    {
-	      if (ppd == NULL || ppd->num_filters == 0)
-		fprintf(stderr, "PAGE: %d %d\n", page, Copies);
-
-              printf("%%%%Page: %d %d\n", page, page);
-	      page ++;
-	      ppdEmit(ppd, stdout, PPD_ORDER_PAGE);
+	      real_page ++;
+	      break;
 	    }
 
-	    start_nup(NumPages);
+          continue;
+        }
+
+        if (!sloworder && NumPages > 0)
+	  end_nup(NumPages - 1);
+
+	if (slowcollate || sloworder)
+	  Pages[NumPages] = ftell(temp);
+
+        if (!sloworder)
+	{
+	  if ((NumPages & (NUp - 1)) == 0)
+	  {
+	    if (ppd == NULL || ppd->num_filters == 0)
+	      fprintf(stderr, "PAGE: %d %d\n", page, Copies);
+
+            printf("%%%%Page: %d %d\n", page, page);
+	    page ++;
+	    ppdEmit(ppd, stdout, PPD_ORDER_PAGE);
 	  }
 
-	  NumPages ++;
+	  start_nup(NumPages);
 	}
+
+	NumPages ++;
+	real_page ++;
       }
       else if (strncmp(line, "%%BeginBinary:", 14) == 0 ||
                (strncmp(line, "%%BeginData:", 12) == 0 &&
-	        strstr(line, "Binary") != NULL))
+	        strstr(line, "ASCII") == NULL && strstr(line, "Hex") == NULL))
       {
        /*
         * Copy binary data...
 	*/
 
         tbytes = atoi(strchr(line, ':') + 1);
+
+	if (!sloworder)
+	  fputs(line, stdout);
+	if (slowcollate || sloworder)
+	  fputs(line, temp);
+
 	while (tbytes > 0)
 	{
-	  nbytes = fread(line, 1, sizeof(line), fp);
+	  if (tbytes > sizeof(line))
+	    nbytes = fread(line, 1, sizeof(line), fp);
+	  else
+	    nbytes = fread(line, 1, tbytes, fp);
+
+          if (nbytes < 1)
+	  {
+	    perror("ERROR: Early end-of-file while reading binary data");
+	    return (1);
+	  }
 
           if (!sloworder)
 	    fwrite(line, 1, nbytes, stdout);
 
           if (slowcollate || sloworder)
-	    fwrite(line, 1, nbytes, stdout);
+	    fwrite(line, 1, nbytes, temp);
 
 	  tbytes -= nbytes;
 	}
       }
-      else if (strcmp(line, "%%Trailer") == 0 && level == 0)
+      else if (strncmp(line, "%%Trailer", 9) == 0 && level == 0)
+      {
+        fputs("DEBUG: Saw Trailer!\n", stderr);
         break;
+      }
       else
       {
         if (!sloworder)
-	  puts(line);
+          fputs(line, stdout);
 
 	if (slowcollate || sloworder)
-	{
 	  fputs(line, temp);
-	  putc('\n', temp);
-	}
       }
 
       if (psgets(line, sizeof(line), fp) == NULL)
@@ -405,16 +524,33 @@ main(int  argc,			/* I - Number of command-line arguments */
 	start_nup(NUp - 1);
         end_nup(NUp - 1);
       }
+
+      if (slowduplex && !(page & 1))
+      {
+       /*
+        * Make sure we have an even number of pages...
+	*/
+
+	if (ppd == NULL || ppd->num_filters == 0)
+	  fprintf(stderr, "PAGE: %d %d\n", page, Copies);
+
+        printf("%%%%Page: %d %d\n", page, page);
+	page ++;
+	ppdEmit(ppd, stdout, PPD_ORDER_PAGE);
+
+	start_nup(NUp - 1);
+	puts("showpage");
+        end_nup(NUp - 1);
+      }
     }
 
     if (slowcollate || sloworder)
     {
       Pages[NumPages] = ftell(temp);
-      page = 1;
 
       if (!sloworder)
       {
-        while (Copies > 0)
+        while (Copies > 1)
 	{
 	  rewind(temp);
 
@@ -441,6 +577,24 @@ main(int  argc,			/* I - Number of command-line arguments */
             end_nup(NUp - 1);
 	  }
 
+	  if (slowduplex && !(page & 1))
+	  {
+	   /*
+            * Make sure we have an even number of pages...
+	    */
+
+	    if (ppd == NULL || ppd->num_filters == 0)
+	      fprintf(stderr, "PAGE: %d 1\n", page);
+
+            printf("%%%%Page: %d %d\n", page, page);
+	    page ++;
+	    ppdEmit(ppd, stdout, PPD_ORDER_PAGE);
+
+	    start_nup(NUp - 1);
+	    puts("showpage");
+            end_nup(NUp - 1);
+	  }
+
 	  Copies --;
 	}
       }
@@ -451,21 +605,25 @@ main(int  argc,			/* I - Number of command-line arguments */
 
         do
 	{
-	  for (page = page_count - 1; page >= 0; page --)
+	  if (slowduplex && (page_count & 1))
+	  {
+            basepage = page_count - 1;
+	  }
+	  else
+	    basepage = page_count - 1 - slowduplex;
+
+	  for (; basepage >= 0; basepage -= 1 + slowduplex)
 	  {
 	    if (ppd == NULL || ppd->num_filters == 0)
-	      fprintf(stderr, "PAGE: %d %d\n", page + 1,
+	      fprintf(stderr, "PAGE: %d %d\n", page,
 	              slowcollate ? 1 : Copies);
 
-            if (slowcollate)
-              printf("%%%%Page: %d %d\n", page + 1,
-	             page_count - page + copy * page_count);
-            else
-	      printf("%%%%Page: %d %d\n", page + 1, page_count - page);
+            printf("%%%%Page: %d %d\n", page, page);
+	    page ++;
 
 	    ppdEmit(ppd, stdout, PPD_ORDER_PAGE);
 
-	    for (subpage = 0, number = page * NUp;
+	    for (subpage = 0, number = basepage * NUp;
 	         subpage < NUp && number < NumPages;
 		 subpage ++, number ++)
 	    {
@@ -480,6 +638,56 @@ main(int  argc,			/* I - Number of command-line arguments */
 	      start_nup(NUp - 1);
               end_nup(NUp - 1);
 	    }
+
+            if (slowduplex)
+	    {
+              if (number < NumPages)
+	      {
+		if (ppd == NULL || ppd->num_filters == 0)
+		  fprintf(stderr, "PAGE: %d %d\n", page,
+	        	  slowcollate ? 1 : Copies);
+
+        	printf("%%%%Page: %d %d\n", page, page);
+		page ++;
+
+		ppdEmit(ppd, stdout, PPD_ORDER_PAGE);
+
+		for (subpage = 0, number = (basepage + 1) * NUp;
+	             subpage < NUp && number < NumPages;
+		     subpage ++, number ++)
+		{
+		  start_nup(number);
+		  fseek(temp, Pages[number], SEEK_SET);
+		  copy_bytes(temp, Pages[number + 1] - Pages[number]);
+		  end_nup(number);
+		}
+
+        	if (number & (NUp - 1))
+		{
+		  start_nup(NUp - 1);
+        	  end_nup(NUp - 1);
+		}
+              }
+	      else
+	      {
+	       /*
+        	* Make sure we have an even number of pages...
+		*/
+
+		if (ppd == NULL || ppd->num_filters == 0)
+		  fprintf(stderr, "PAGE: %d %d\n", page, slowcollate ? 1 : Copies);
+
+        	printf("%%%%Page: %d %d\n", page, page);
+		page ++;
+		ppdEmit(ppd, stdout, PPD_ORDER_PAGE);
+
+		start_nup(NUp - 1);
+		puts("showpage");
+        	end_nup(NUp - 1);
+
+        	basepage = page_count - 1;
+	      }
+	    }
 	  }
 
 	  copy ++;
@@ -492,8 +700,18 @@ main(int  argc,			/* I - Number of command-line arguments */
     * Copy the trailer, if any...
     */
 
-    while ((nbytes = fread(line, 1, sizeof(line), fp)) > 0)
-      fwrite(line, 1, nbytes, stdout);
+    while (psgets(line, sizeof(line), fp) != NULL)
+    {
+      if (strcmp(line, "\004") != 0)
+        fputs(line, stdout);
+
+      if (strncmp(line, "%%EOF", 5) == 0)
+      {
+        fputs("DEBUG: Saw EOF!\n", stderr);
+        saweof = 1;
+	break;
+      }
+    }
   }
   else
   {
@@ -506,17 +724,19 @@ main(int  argc,			/* I - Number of command-line arguments */
 
     ppdEmit(ppd, stdout, PPD_ORDER_PAGE);
 
-    while (psgets(line, sizeof(line), fp) != NULL)
+    saweof = 1;
+
+    while ((nbytes = fread(line, 1, sizeof(line), fp)) > 0)
     {
-      puts(line);
+      fwrite(line, 1, nbytes, stdout);
 
       if (slowcollate)
-      {
-	fputs(line, temp);
-	putc('\n', temp);
-      }
+	fwrite(line, 1, nbytes, temp);
     }
 
+    if (UseESPsp)
+      puts("ESPshowpage");
+
     if (slowcollate)
     {
       while (Copies > 1)
@@ -527,17 +747,28 @@ main(int  argc,			/* I - Number of command-line arguments */
         ppdEmit(ppd, stdout, PPD_ORDER_PAGE);
 	rewind(temp);
 	copy_bytes(temp, 0);
+	Copies --;
+
+        if (UseESPsp)
+          puts("ESPshowpage");
       }
     }
   }
 
+ /*
+  * Send %%EOF if needed...
+  */
+
+  if (!saweof)
+    puts("%%EOF");
+
  /*
   * End the job with the appropriate JCL command or CTRL-D otherwise.
   */
 
   if (ppd != NULL && ppd->jcl_end)
     fputs(ppd->jcl_end, stdout);
-  else
+  else if (ppd != NULL && ppd->num_filters == 0)
     putchar(0x04);
 
  /*
@@ -577,9 +808,9 @@ check_range(int page)	/* I - Page number */
     * See if we only print even or odd pages...
     */
 
-    if (strcasecmp(PageSet, "even") == 0 && (page & 1))
+    if (strcasecmp(PageSet, "even") == 0 && ((page - 1) % (NUp << 1)) <  NUp)
       return (0);
-    if (strcasecmp(PageSet, "odd") == 0 && !(page & 1))
+    if (strcasecmp(PageSet, "odd")  == 0 && ((page - 1) % (NUp << 1)) >= NUp)
       return (0);
   }
 
@@ -640,7 +871,7 @@ copy_bytes(FILE   *fp,		/* I - File to read from */
 
   while (nleft > 0 || length == 0)
   {
-    if (nleft > sizeof(buffer))
+    if (nleft > sizeof(buffer) || length == 0)
       nbytes = sizeof(buffer);
     else
       nbytes = nleft;
@@ -663,18 +894,48 @@ static void
 end_nup(int number)	/* I - Page number */
 {
   if (Flip || Orientation || NUp > 1)
-    puts("ESPsave restore");
+    puts("userdict /ESPsave get restore");
 
   switch (NUp)
   {
-    case 2 :
-	if ((number & 1) == 1)
+    case 1 :
+	if (UseESPsp)
+	{
+	  WriteLabels(Orientation);
           puts("ESPshowpage");
+	}
+	break;
+
+    case 2 :
+	if ((number & 1) == 1 && UseESPsp)
+	{
+	  if (Orientation & 1)
+	  {
+	   /*
+	    * Rotate the labels back to portrait...
+	    */
+
+	    WriteLabels(Orientation - 1);
+	  }
+	  else
+	  {
+	   /*
+	    * Rotate the labels to landscape...
+	    */
+
+	    WriteLabels(Orientation + 1);
+	  }
+
+          puts("ESPshowpage");
+	}
         break;
 
     case 4 :
-	if ((number & 3) == 3)
+	if ((number & 3) == 3 && UseESPsp)
+	{
+	  WriteLabels(Orientation);
           puts("ESPshowpage");
+	}
         break;
   }
 }
@@ -701,6 +962,7 @@ psgets(char   *buf,	/* I - Buffer to read into */
 
   len --;
   bufptr = buf;
+  ch     = EOF;
 
   while ((bufptr - buf) < len)
   {
@@ -717,6 +979,7 @@ psgets(char   *buf,	/* I - Buffer to read into */
       if (ch != EOF && ch != 0x0a)
         ungetc(ch, fp);	/* Nope, save it for later... */
 
+      ch = 0x0a;
       break;
     }
     else if (ch == 0x0a)
@@ -725,6 +988,13 @@ psgets(char   *buf,	/* I - Buffer to read into */
       *bufptr++ = ch;
   }
 
+ /*
+  * Add a trailing newline if it is there...
+  */
+
+  if (ch == '\n')
+    *bufptr++ = '\n';
+
  /*
   * Nul-terminate the string and return it (or NULL for EOF).
   */
@@ -752,24 +1022,29 @@ start_nup(int number)	/* I - Page number */
 
 
   if (Flip || Orientation || NUp > 1)
-    puts("/ESPsave save def");
+    puts("userdict /ESPsave save put");
 
   if (Flip)
-    printf("%.0f 0 translate -1 1 scale\n", PageWidth);
+    printf("%.1f 0.0 translate -1 1 scale\n", PageWidth);
 
   pw = PageRight - PageLeft;
   pl = PageTop - PageBottom;
 
+  fprintf(stderr, "DEBUG: pw = %.1f, pl = %.1f\n", pw, pl);
+  fprintf(stderr, "DEBUG: PageLeft = %.1f, PageRight = %.1f\n", PageLeft, PageRight);
+  fprintf(stderr, "DEBUG: PageTop = %.1f, PageBottom = %.1f\n", PageTop, PageBottom);
+  fprintf(stderr, "DEBUG: PageWidth = %.1f, PageLength = %.1f\n", PageWidth, PageLength);
+
   switch (Orientation)
   {
     case 1 : /* Landscape */
-        printf("%.0f 0 translate 90 rotate\n", PageLength);
+        printf("%.1f 0.0 translate 90 rotate\n", PageLength);
         break;
     case 2 : /* Reverse Portrait */
-        printf("%.0f %.0f translate 180 rotate\n", PageWidth, PageLength);
+        printf("%.1f %.1f translate 180 rotate\n", PageWidth, PageLength);
         break;
     case 3 : /* Reverse Landscape */
-        printf("0 %.0f translate -90 rotate\n", PageWidth);
+        printf("0.0 %.1f translate -90 rotate\n", PageWidth);
         break;
   }
 
@@ -809,28 +1084,28 @@ start_nup(int number)	/* I - Page number */
         }
 
         if (Duplex && (number & 2))
-	  printf("%.0f %.0f translate\n", PageWidth - PageRight, PageBottom);
+	  printf("%.1f %.1f translate\n", PageWidth - PageRight, PageBottom);
 	else
-	  printf("%.0f %.0f translate\n", PageLeft, PageBottom);
+	  printf("%.1f %.1f translate\n", PageLeft, PageBottom);
 
         if (Orientation & 1)
 	{
-          printf("0 %.0f translate -90 rotate\n", pl);
-          printf("%.0f %.0f translate %.3f %.3f scale\n",
-                 ty, tx + l * x, w / pw, l / pl);
+          printf("0.0 %.1f translate -90 rotate\n", pl);
+          printf("%.1f %.1f translate %.3f %.3f scale\n",
+                 ty, tx + l * x, w / PageWidth, l / PageLength);
         }
         else
 	{
-          printf("%.0f 0 translate 90 rotate\n", pw);
-          printf("%.0f %.0f translate %.3f %.3f scale\n",
-                 tx + w * x, ty, w / pw, l / pl);
+          printf("%.1f 0.0 translate 90 rotate\n", pw);
+          printf("%.1f %.1f translate %.3f %.3f scale\n",
+                 tx + w * x, ty, w / PageWidth, l / PageLength);
         }
 
 	printf("newpath\n"
-               "0 0 moveto\n"
-               "%.0f 0 lineto\n"
-               "%.0f %.0f lineto\n"
-               "0 %.0f lineto\n"
+               "0.0 0.0 moveto\n"
+               "%.1f 0.0 lineto\n"
+               "%.1f %.1f lineto\n"
+               "0.0 %.1f lineto\n"
                "closepath clip newpath\n",
                PageWidth, PageWidth, PageLength, PageLength);
         break;
@@ -849,17 +1124,17 @@ start_nup(int number)	/* I - Page number */
 	}
 
         if (Duplex && (number & 4))
-	  printf("%.0f %.0f translate\n", PageWidth - PageRight, PageBottom);
+	  printf("%.1f %.1f translate\n", PageWidth - PageRight, PageBottom);
 	else
-	  printf("%.0f %.0f translate\n", PageLeft, PageBottom);
+	  printf("%.1f %.1f translate\n", PageLeft, PageBottom);
 
-	printf("%.0f %.0f translate %.3f %.3f scale\n", x * w, y * l,
+	printf("%.1f %.1f translate %.3f %.3f scale\n", x * w, y * l,
 	       w / PageWidth, l / PageLength);
         printf("newpath\n"
-               "0 0 moveto\n"
-               "%.0f 0 lineto\n"
-               "%.0f %.0f lineto\n"
-               "0 %.0f lineto\n"
+               "0.0 0.0 moveto\n"
+               "%.1f 0.0 lineto\n"
+               "%.1f %.1f lineto\n"
+               "0.0 %.1f lineto\n"
                "closepath clip newpath\n",
                PageWidth, PageWidth, PageLength, PageLength);
         break;
@@ -868,5 +1143,5 @@ start_nup(int number)	/* I - Page number */
 
 
 /*
- * End of "$Id: pstops.c,v 1.32 1999/12/07 19:05:45 mike Exp $".
+ * End of "$Id: pstops.c,v 1.72 2002/01/23 22:30:42 mike Exp $".
  */
