@@ -1,9 +1,9 @@
 /*
- * "$Id: textcommon.c,v 1.10 1999/08/04 16:12:29 mike Exp $"
+ * "$Id: textcommon.c,v 1.20 2002/01/17 21:39:26 mike Exp $"
  *
  *   Common text filter routines for the Common UNIX Printing System (CUPS).
  *
- *   Copyright 1997-1999 by Easy Software Products.
+ *   Copyright 1997-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -49,10 +49,18 @@ int	WrapLines = 1,		/* Wrap text in lines */
 	Copies = 1;		/* Number of copies */
 lchar_t	**Page = NULL;		/* Page characters */
 int	NumPages = 0;		/* Number of pages in document */
-int	CharsPerInch = 10;	/* Number of character columns per inch */
-int	LinesPerInch = 6;	/* Number of lines per inch */
+float	CharsPerInch = 10;	/* Number of character columns per inch */
+float	LinesPerInch = 6;	/* Number of lines per inch */
 int	UTF8 = 0;		/* Use UTF-8 encoding? */
-char	*Keywords[] =		/* List of known keywords... */
+int	NumKeywords = 0;	/* Number of known keywords */
+char	**Keywords = NULL;	/* List of known keywords */
+
+
+/*
+ * Local globals...
+ */
+
+static char *code_keywords[] =	/* List of known C/C++ keywords... */
 	{
 	  "and",
 	  "and_eq",
@@ -120,6 +128,344 @@ char	*Keywords[] =		/* List of known keywords... */
 	  "while",
 	  "xor",
 	  "xor_eq"
+	},
+	*sh_keywords[] =	/* List of known Boure/Korn/zsh/bash keywords... */
+	{
+	  "alias",
+	  "bg",
+	  "break",
+	  "case",
+	  "cd",
+	  "command",
+	  "continue",
+	  "do",
+	  "done",
+	  "echo",
+	  "elif",
+	  "else",
+	  "esac",
+	  "eval",
+	  "exec",
+	  "exit",
+	  "export",
+	  "fc",
+	  "fg",
+	  "fi",
+	  "for",
+	  "function",
+	  "getopts",
+	  "if",
+	  "in",
+	  "jobs",
+	  "kill",
+	  "let",
+	  "limit",
+	  "newgrp",
+	  "print",
+	  "pwd",
+	  "read",
+	  "readonly",
+	  "return",
+	  "select",
+	  "set",
+	  "shift",
+	  "test",
+	  "then",
+	  "time",
+	  "times",
+	  "trap",
+	  "typeset",
+	  "ulimit",
+	  "umask",
+	  "unalias",
+	  "unlimit",
+	  "unset",
+	  "until",
+	  "wait",
+	  "whence"
+	  "while",
+	},
+	*csh_keywords[] =	/* List of known csh/tcsh keywords... */
+	{
+	  "alias",
+	  "aliases",
+	  "bg",
+	  "bindkey",
+	  "break",
+	  "breaksw",
+	  "builtins",
+	  "case",
+	  "cd",
+	  "chdir",
+	  "complete",
+	  "continue",
+	  "default",
+	  "dirs",
+	  "echo",
+	  "echotc",
+	  "else",
+	  "end",
+	  "endif",
+	  "eval",
+	  "exec",
+	  "exit",
+	  "fg",
+	  "foreach",
+	  "glob",
+	  "goto",
+	  "history",
+	  "if",
+	  "jobs",
+	  "kill",
+	  "limit",
+	  "login",
+	  "logout",
+	  "ls",
+	  "nice",
+	  "nohup",
+	  "notify",
+	  "onintr",
+	  "popd",
+	  "pushd",
+	  "pwd",
+	  "rehash",
+	  "repeat",
+	  "set",
+	  "setenv",
+	  "settc",
+	  "shift",
+	  "source",
+	  "stop",
+	  "suspend",
+	  "switch",
+	  "telltc",
+	  "then",
+	  "time",
+	  "umask",
+	  "unalias",
+	  "unbindkey",
+	  "unhash",
+	  "unlimit",
+	  "unset",
+	  "unsetenv",
+	  "wait",
+	  "where",
+	  "which",
+	  "while"
+	},
+	*perl_keywords[] =	/* List of known perl keywords... */
+	{
+	  "abs",
+	  "accept",
+	  "alarm",
+	  "and",
+	  "atan2",
+	  "bind",
+	  "binmode",
+	  "bless",
+	  "caller",
+	  "chdir",
+	  "chmod",
+	  "chomp",
+	  "chop",
+	  "chown",
+	  "chr",
+	  "chroot",
+	  "closdir",
+	  "close",
+	  "connect",
+	  "continue",
+	  "cos",
+	  "crypt",
+	  "dbmclose",
+	  "dbmopen",
+	  "defined",
+	  "delete",
+	  "die",
+	  "do",
+	  "dump",
+	  "each",
+	  "else",
+	  "elsif",
+	  "endgrent",
+	  "endhostent",
+	  "endnetent",
+	  "endprotoent",
+	  "endpwent",
+	  "endservent",
+	  "eof",
+	  "eval",
+	  "exec",
+	  "exists",
+	  "exit",
+	  "exp",
+	  "fcntl",
+	  "fileno",
+	  "flock",
+	  "for",
+	  "foreach",
+	  "fork",
+	  "format",
+	  "formline",
+	  "getc",
+	  "getgrent",
+	  "getgrgid",
+	  "getgrnam",
+	  "gethostbyaddr",
+	  "gethostbyname",
+	  "gethostent",
+	  "getlogin",
+	  "getnetbyaddr",
+	  "getnetbyname",
+	  "getnetent",
+	  "getpeername",
+	  "getpgrp",
+	  "getppid",
+	  "getpriority",
+	  "getprotobyname",
+	  "getprotobynumber",
+	  "getprotoent",
+	  "getpwent",
+	  "getpwnam",
+	  "getpwuid",
+	  "getservbyname",
+	  "getservbyport",
+	  "getservent",
+	  "getsockname",
+	  "getsockopt",
+	  "glob",
+	  "gmtime",
+	  "goto",
+	  "grep",
+	  "hex",
+	  "if",
+	  "import",
+	  "index",
+	  "int",
+	  "ioctl",
+	  "join",
+	  "keys",
+	  "kill",
+	  "last",
+	  "lc",
+	  "lcfirst",
+	  "length",
+	  "link",
+	  "listen",
+	  "local",
+	  "localtime",
+	  "log",
+	  "lstat",
+	  "map",
+	  "mkdir",
+	  "msgctl",
+	  "msgget",
+	  "msgrcv",
+	  "msgsend",
+	  "my",
+	  "next",
+	  "no",
+	  "not",
+	  "oct",
+	  "open",
+	  "opendir",
+	  "or",
+	  "ord",
+	  "pack",
+	  "package",
+	  "pipe",
+	  "pop",
+	  "pos",
+	  "print",
+	  "printf",
+	  "push",
+	  "quotemeta",
+	  "rand",
+	  "read",
+	  "readdir",
+	  "readlink",
+	  "recv",
+	  "redo",
+	  "ref",
+	  "rename",
+	  "require",
+	  "reset",
+	  "return",
+	  "reverse",
+	  "rewinddir",
+	  "rindex",
+	  "rmdir",
+	  "scalar",
+	  "seek",
+	  "seekdir",
+	  "select",
+	  "semctl",
+	  "semget",
+	  "semop",
+	  "send",
+	  "setgrent",
+	  "sethostent",
+	  "setnetent",
+	  "setpgrp",
+	  "setpriority",
+	  "setprotoent",
+	  "setpwent",
+	  "setservent",
+	  "setsockopt",
+	  "shift",
+	  "shmctl",
+	  "shmget",
+	  "shmread",
+	  "shmwrite",
+	  "shutdown",
+	  "sin",
+	  "sleep",
+	  "socket",
+	  "socketpair",
+	  "sort",
+	  "splice",
+	  "split",
+	  "sprintf",
+	  "sqrt",
+	  "srand",
+	  "stat",
+	  "study",
+	  "sub",
+	  "substr",
+	  "symlink",
+	  "syscall",
+	  "sysread",
+	  "sysseek",
+	  "system",
+	  "syswrite",
+	  "tell",
+	  "telldir",
+	  "tie",
+	  "tied",
+	  "time",
+	  "times"
+	  "times",
+	  "truncate",
+	  "uc",
+	  "ucfirst",
+	  "umask",
+	  "undef",
+	  "unless",
+	  "unlink",
+	  "unpack",
+	  "unshift",
+	  "untie",
+	  "until",
+	  "use",
+	  "utime",
+	  "values",
+	  "vec",
+	  "wait",
+	  "waitpid",
+	  "wantarray",
+	  "warn",
+	  "while",
+	  "write"
 	};
 
 
@@ -136,9 +482,9 @@ static int	getutf8(FILE *fp);
  */
 
 int				/* O - Exit status */
-TextMain(char *name,		/* I - Name of filter */
-         int  argc,		/* I - Number of command-line arguments */
-         char *argv[])		/* I - Command-line arguments */
+TextMain(const char *name,	/* I - Name of filter */
+         int        argc,	/* I - Number of command-line arguments */
+         char       *argv[])	/* I - Command-line arguments */
 {
   FILE		*fp;		/* Print file */
   ppd_file_t	*ppd;		/* PPD file */
@@ -159,6 +505,16 @@ TextMain(char *name,		/* I - Name of filter */
   int		cstring;	/* Inside a C string */
 
 
+ /*
+  * Make sure status messages are not buffered...
+  */
+
+  setbuf(stderr, NULL);
+
+ /*
+  * Check command-line...
+  */
+
   if (argc < 6 || argc > 7)
   {
     fprintf(stderr, "ERROR: %s job-id user title copies options [file]\n",
@@ -195,17 +551,46 @@ TextMain(char *name,		/* I - Name of filter */
 
   if ((val = cupsGetOption("prettyprint", num_options, options)) != NULL)
   {
-    PrettyPrint  = 1;
     PageLeft     = 72.0f;
     PageRight    = PageWidth - 36.0f;
     PageBottom   = PageBottom > 36.0f ? PageBottom : 36.0f;
     PageTop      = PageLength - 36.0f;
     CharsPerInch = 12;
     LinesPerInch = 8;
+
+    if ((val = getenv("CONTENT_TYPE")) == NULL)
+    {
+      PrettyPrint = PRETTY_CODE;
+      NumKeywords = sizeof(code_keywords) / sizeof(code_keywords[0]);
+      Keywords    = code_keywords;
+    }
+    else if (strcasecmp(val, "application/x-cshell") == 0)
+    {
+      PrettyPrint = PRETTY_SHELL;
+      NumKeywords = sizeof(csh_keywords) / sizeof(csh_keywords[0]);
+      Keywords    = csh_keywords;
+    }
+    else if (strcasecmp(val, "application/x-perl") == 0)
+    {
+      PrettyPrint = PRETTY_PERL;
+      NumKeywords = sizeof(perl_keywords) / sizeof(perl_keywords[0]);
+      Keywords    = perl_keywords;
+    }
+    else if (strcasecmp(val, "application/x-shell") == 0)
+    {
+      PrettyPrint = PRETTY_SHELL;
+      NumKeywords = sizeof(sh_keywords) / sizeof(sh_keywords[0]);
+      Keywords    = sh_keywords;
+    }
+    else
+    {
+      PrettyPrint = PRETTY_CODE;
+      NumKeywords = sizeof(code_keywords) / sizeof(code_keywords[0]);
+      Keywords    = code_keywords;
+    }
   }
 
-  if ((ppd = SetCommonOptions(num_options, options, 1)) != NULL)
-    ppdClose(ppd);
+  ppd = SetCommonOptions(num_options, options, 1);
 
   WrapLines = cupsGetOption("nowrap", num_options, options) == NULL;
 
@@ -213,22 +598,24 @@ TextMain(char *name,		/* I - Name of filter */
     PageColumns = atoi(val);
 
   if ((val = cupsGetOption("cpi", num_options, options)) != NULL)
-    CharsPerInch = atoi(val);
+    CharsPerInch = atof(val);
 
   if ((val = cupsGetOption("lpi", num_options, options)) != NULL)
-    LinesPerInch = atoi(val);
+    LinesPerInch = atof(val);
 
   if ((val = cupsGetOption("prettyprint", num_options, options)) != NULL)
     PageTop -= 216.0f / LinesPerInch;
 
   Copies = atoi(argv[4]);
 
-  WriteProlog(argv[3], argv[2], ppd);
+  WriteProlog(argv[3], argv[2], getenv("CLASSIFICATION"),
+              cupsGetOption("page-label", num_options, options), ppd);
 
  /*
   * Read text from the specified source and print it...
   */
 
+  lastch       = 0;
   column       = 0;
   line         = 0;
   page_column  = 0;
@@ -270,8 +657,8 @@ TextMain(char *name,		/* I - Name of filter */
 	    *keyptr = '\0';
 	    keyptr  = keyword;
 
-	    if (bsearch(&keyptr, Keywords, sizeof(Keywords) / sizeof(Keywords[0]),
-	                sizeof(Keywords[0]), compare_keywords))
+	    if (bsearch(&keyptr, Keywords, NumKeywords, sizeof(char *),
+	                compare_keywords))
             {
 	     /*
 	      * Put keywords in boldface...
@@ -316,8 +703,8 @@ TextMain(char *name,		/* I - Name of filter */
 	    *keyptr = '\0';
 	    keyptr  = keyword;
 
-	    if (bsearch(&keyptr, Keywords, sizeof(Keywords) / sizeof(Keywords[0]),
-	                sizeof(Keywords[0]), compare_keywords))
+	    if (bsearch(&keyptr, Keywords, NumKeywords, sizeof(char *),
+	                compare_keywords))
             {
 	     /*
 	      * Put keywords in boldface...
@@ -370,8 +757,8 @@ TextMain(char *name,		/* I - Name of filter */
 	    *keyptr = '\0';
 	    keyptr  = keyword;
 
-	    if (bsearch(&keyptr, Keywords, sizeof(Keywords) / sizeof(Keywords[0]),
-	                sizeof(Keywords[0]), compare_keywords))
+	    if (bsearch(&keyptr, Keywords, NumKeywords, sizeof(char *),
+	                compare_keywords))
             {
 	     /*
 	      * Put keywords in boldface...
@@ -480,8 +867,8 @@ TextMain(char *name,		/* I - Name of filter */
 	      keyptr  = keyword;
 
 	      if (!(attr & ATTR_ITALIC) &&
-	          bsearch(&keyptr, Keywords, sizeof(Keywords) / sizeof(Keywords[0]),
-	                  sizeof(Keywords[0]), compare_keywords))
+	          bsearch(&keyptr, Keywords, NumKeywords, sizeof(char *),
+	                  compare_keywords))
               {
 	       /*
 	        * Put keywords in boldface...
@@ -517,7 +904,8 @@ TextMain(char *name,		/* I - Name of filter */
 	      cstring = -1;
               attr    |= ATTR_BLUE;
 	    }
-            else if (ch == '*' && lastch == '/' && !cstring)
+            else if (ch == '*' && lastch == '/' && !cstring &&
+	             PrettyPrint != PRETTY_SHELL)
 	    {
 	     /*
 	      * Start a C-style comment...
@@ -526,7 +914,8 @@ TextMain(char *name,		/* I - Name of filter */
 	      ccomment = 1;
 	      attr     |= ATTR_ITALIC | ATTR_GREEN;
 	    }
-	    else if (ch == '/' && lastch == '/' && !cstring)
+	    else if (ch == '/' && lastch == '/' && !cstring &&
+	             PrettyPrint == PRETTY_CODE)
 	    {
 	     /*
 	      * Start a C++-style comment...
@@ -534,7 +923,16 @@ TextMain(char *name,		/* I - Name of filter */
 
 	      attr |= ATTR_ITALIC | ATTR_GREEN;
 	    }
-	    else if (ch == '#' && column == 0 && !ccomment && !cstring)
+	    else if (ch == '#' && !cstring && PrettyPrint != PRETTY_CODE)
+	    {
+	     /*
+	      * Start a shell-style comment...
+	      */
+
+	      attr |= ATTR_ITALIC | ATTR_GREEN;
+	    }
+	    else if (ch == '#' && column == 0 && !ccomment && !cstring &&
+	             PrettyPrint == PRETTY_CODE)
 	    {
 	     /*
 	      * Start a preprocessor command...
@@ -572,10 +970,19 @@ TextMain(char *name,		/* I - Name of filter */
 
             if (PrettyPrint)
               Page[line][i].attr = attr;
+	    else if (ch == ' ' && Page[line][i].ch)
+	      ch = Page[line][i].ch;
             else if (ch == Page[line][i].ch)
               Page[line][i].attr |= ATTR_BOLD;
             else if (Page[line][i].ch == '_')
               Page[line][i].attr |= ATTR_UNDERLINE;
+            else if (ch == '_')
+	    {
+              Page[line][i].attr |= ATTR_UNDERLINE;
+
+              if (Page[line][i].ch)
+	        ch = Page[line][i].ch;
+	    }
 	    else
               Page[line][i].attr = attr;
 
@@ -591,16 +998,16 @@ TextMain(char *name,		/* I - Name of filter */
 	      * Highlight curley braces...
 	      */
 
-	      Page[line][i].attr |= ATTR_BOLD;
+	      Page[line][column].attr |= ATTR_BOLD;
 	    }
 	    else if ((ch == '/' || ch == '*') && lastch == '/' &&
-	             column < ColumnWidth)
+	             column < ColumnWidth && PrettyPrint != PRETTY_SHELL)
 	    {
 	     /*
 	      * Highlight first comment character...
 	      */
 
-	      Page[line][i - 1].attr = attr;
+	      Page[line][column - 1].attr = attr;
 	    }
 	    else if (ch == '\"' && lastch != '\\' && !ccomment && cstring > 0)
 	    {
@@ -649,6 +1056,9 @@ TextMain(char *name,		/* I - Name of filter */
 
   WriteEpilogue();
 
+  if (ppd != NULL)
+    ppdClose(ppd);
+
   return (0);
 }
 
@@ -741,5 +1151,5 @@ getutf8(FILE *fp)	/* I - File to read from */
 
 
 /*
- * End of "$Id: textcommon.c,v 1.10 1999/08/04 16:12:29 mike Exp $".
+ * End of "$Id: textcommon.c,v 1.20 2002/01/17 21:39:26 mike Exp $".
  */
