@@ -1,9 +1,9 @@
 /*
- * "$Id: lpadmin.c,v 1.10 1999/10/10 15:41:15 mike Exp $"
+ * "$Id: lpadmin.c,v 1.28 2002/01/02 17:59:19 mike Exp $"
  *
  *   "lpadmin" command for the Common UNIX Printing System (CUPS).
  *
- *   Copyright 1997-1999 by Easy Software Products.
+ *   Copyright 1997-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -33,6 +33,10 @@
  *   set_printer_file()          - Set the interface script or PPD file.
  *   set_printer_info()          - Set the printer description string.
  *   set_printer_location()      - Set the printer location string.
+ *   set_printer_model()         - Set the driver model file.
+ *   set_printer_options()       - Set the printer options.
+ *   validate_name()             - Make sure the printer name only contains
+ *                                 letters, numbers, and the underscore...
  */
 
 /*
@@ -43,6 +47,7 @@
 #include <stdlib.h>
 #include <ctype.h>
 #include <cups/cups.h>
+#include <cups/string.h>
 #include <cups/language.h>
 #include <cups/debug.h>
 #include <config.h>
@@ -60,10 +65,14 @@ static void	default_printer(http_t *, char *);
 static void	delete_printer(http_t *, char *);
 static void	delete_printer_from_class(http_t *, char *, char *);
 static void	enable_printer(http_t *, char *);
+static char	*get_line(char *, int, FILE *fp);
 static void	set_printer_device(http_t *, char *, char *);
 static void	set_printer_file(http_t *, char *, char *);
 static void	set_printer_info(http_t *, char *, char *);
 static void	set_printer_location(http_t *, char *, char *);
+static void	set_printer_model(http_t *, char *, char *);
+static void	set_printer_options(http_t *, char *, int, cups_option_t *);
+static int	validate_name(const char *);
 
 
 /*
@@ -71,29 +80,41 @@ static void	set_printer_location(http_t *, char *, char *);
  */
 
 int
-main(int  argc,		/* I - Number of command-line arguments */
-     char *argv[])	/* I - Command-line arguments */
+main(int  argc,			/* I - Number of command-line arguments */
+     char *argv[])		/* I - Command-line arguments */
 {
-  int	i;		/* Looping var */
-  http_t *http;		/* Connection to server */
-  char	*printer,	/* Destination printer */
-	*host,		/* Pointer to hostname */
-	filename[1024];	/* Model filename */
+  int		i;		/* Looping var */
+  http_t	*http;		/* Connection to server */
+  char		*printer,	/* Destination printer */
+		*pclass,	/* Printer class name */
+		*val;		/* Pointer to allow/deny value */
+  int		num_options;	/* Number of options */
+  cups_option_t	*options;	/* Options */
+  http_encryption_t encryption;	/* Encryption? */
 
 
-  if ((http = httpConnect(cupsServer(), ippPort())) == NULL)
-  {
-    fputs("lpadmin: Unable to contact server!\n", stderr);
-    return (1);
-  }
-
-  printer = NULL;
+  http        = NULL;
+  printer     = NULL;
+  num_options = 0;
+  options     = NULL;
+  encryption  = cupsEncryption();
 
   for (i = 1; i < argc; i ++)
     if (argv[i][0] == '-')
       switch (argv[i][1])
       {
         case 'c' : /* Add printer to class */
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
 	    if (printer == NULL)
 	    {
 	      fputs("lpadmin: Unable to add a printer to the class:\n", stderr);
@@ -102,43 +123,83 @@ main(int  argc,		/* I - Number of command-line arguments */
 	    }
 
 	    if (argv[i][2])
-	      add_printer_to_class(http, printer, argv[i] + 2);
+	      pclass = argv[i] + 2;
 	    else
 	    {
 	      i ++;
-	      add_printer_to_class(http, printer, argv[i]);
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected class name after \'-c\' option!\n", stderr);
+		return (1);
+	      }
+
+	      pclass = argv[i];
 	    }
+
+            if (!validate_name(pclass))
+	    {
+	      fputs("lpadmin: Class name can only contain letters, numbers, and the underscore!\n", stderr);
+	      return (1);
+	    }
+
+	    add_printer_to_class(http, printer, pclass);
 	    break;
 
         case 'd' : /* Set as default destination */
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
 	    if (argv[i][2])
 	      printer = argv[i] + 2;
 	    else
 	    {
 	      i ++;
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected printer name after \'-d\' option!\n", stderr);
+		return (1);
+	      }
+
 	      printer = argv[i];
 	    }
 
+            if (!validate_name(printer))
+	    {
+	      fputs("lpadmin: Printer name can only contain letters, numbers, and the underscore!\n", stderr);
+	      return (1);
+	    }
+
             default_printer(http, printer);
 	    i = argc;
 	    break;
 
         case 'h' : /* Connect to host */
-	    httpClose(http);
+	    if (http)
+	      httpClose(http);
 
 	    if (argv[i][2] != '\0')
-	      http = httpConnect(argv[i] + 2, ippPort());
+	      http = httpConnectEncrypt(argv[i] + 2, ippPort(), encryption);
 	    else
 	    {
 	      i ++;
 
 	      if (i >= argc)
 	      {
-	        fputs("Error: need hostname after \'-h\' option!\n", stderr);
+	        fputs("lpadmin: Expected hostname after \'-h\' option!\n", stderr);
 		return (1);
               }
 	      else
-		http = httpConnect(argv[i], ippPort());
+		http = httpConnectEncrypt(argv[i], ippPort(), encryption);
 	    }
 
 	    if (http == NULL)
@@ -146,9 +207,22 @@ main(int  argc,		/* I - Number of command-line arguments */
 	      perror("lpadmin: Unable to connect to server");
 	      return (1);
 	    }
+	    else
+	      cupsSetServer(http->hostname);
 	    break;
 
         case 'i' : /* Use the specified interface script */
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
 	    if (printer == NULL)
 	    {
 	      fputs("lpadmin: Unable to set the interface script:\n", stderr);
@@ -161,6 +235,13 @@ main(int  argc,		/* I - Number of command-line arguments */
 	    else
 	    {
 	      i ++;
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected interface after \'-i\' option!\n", stderr);
+		return (1);
+	      }
+
 	      set_printer_file(http, printer, argv[i]);
 	    }
 	    break;
@@ -168,15 +249,44 @@ main(int  argc,		/* I - Number of command-line arguments */
         case 'E' : /* Enable the printer */
 	    if (printer == NULL)
 	    {
-	      fputs("lpadmin: Unable to enable the printer:\n", stderr);
-	      fputs("         You must specify a printer name first!\n", stderr);
-	      return (1);
+#ifdef HAVE_LIBSSL
+	      cupsSetEncryption(encryption = HTTP_ENCRYPT_REQUIRED);
+
+	      if (http)
+		httpEncryption(http, encryption);
+#else
+              fprintf(stderr, "%s: Sorry, no encryption support compiled in!\n",
+	              argv[0]);
+#endif /* HAVE_LIBSSL */
+	      break;
 	    }
 
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
             enable_printer(http, printer);
             break;
 
         case 'm' : /* Use the specified standard script/PPD file */
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
 	    if (printer == NULL)
 	    {
 	      fputs("lpadmin: Unable to set the interface script or PPD file:\n", stderr);
@@ -185,43 +295,84 @@ main(int  argc,		/* I - Number of command-line arguments */
 	    }
 
 	    if (argv[i][2])
-	      snprintf(filename, sizeof(filename), CUPS_DATADIR "/model/%s", argv[i] + 2);
+	      set_printer_model(http, printer, argv[i] + 2);
 	    else
 	    {
 	      i ++;
-	      snprintf(filename, sizeof(filename), CUPS_DATADIR "/model/%s", argv[i]);
-	    }
 
-            set_printer_file(http, printer, filename);
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected model after \'-m\' option!\n", stderr);
+		return (1);
+	      }
+
+	      set_printer_model(http, printer, argv[i]);
+	    }
+	    break;
+
+        case 'o' : /* Set option */
+	    if (argv[i][2])
+	      num_options = cupsParseOptions(argv[i] + 2, num_options, &options);
+	    else
+	    {
+	      i ++;
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected name=value after \'-o\' option!\n", stderr);
+		return (1);
+	      }
+
+	      num_options = cupsParseOptions(argv[i], num_options, &options);
+	    }
 	    break;
 
         case 'p' : /* Add/modify a printer */
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
 	    if (argv[i][2])
 	      printer = argv[i] + 2;
 	    else
 	    {
 	      i ++;
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected printer after \'-p\' option!\n", stderr);
+		return (1);
+	      }
+
 	      printer = argv[i];
 	    }
 
-	    if ((host = strchr(printer, '@')) != NULL)
+            if (!validate_name(printer))
 	    {
-	     /*
-	      * printer@host - reconnect to the named host...
-	      */
-
-	      httpClose(http);
-
-              *host++ = '\0';
-              if ((http = httpConnect(host, ippPort())) == NULL)
-	      {
-		perror("lpadmin: Unable to connect to server");
-		return (1);
-	      }
+	      fputs("lpadmin: Printer name can only contain letters, numbers, and the underscore!\n", stderr);
+	      return (1);
 	    }
 	    break;
 
         case 'r' : /* Remove printer from class */
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
 	    if (printer == NULL)
 	    {
 	      fputs("lpadmin: Unable to remove a printer from the class:\n", stderr);
@@ -230,15 +381,71 @@ main(int  argc,		/* I - Number of command-line arguments */
 	    }
 
 	    if (argv[i][2])
-	      delete_printer_from_class(http, printer, argv[i] + 2);
+	      pclass = argv[i] + 2;
 	    else
 	    {
 	      i ++;
-	      delete_printer_from_class(http, printer, argv[i]);
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected class after \'-r\' option!\n", stderr);
+		return (1);
+	      }
+
+	      pclass = argv[i];
+	    }
+
+            if (!validate_name(pclass))
+	    {
+	      fputs("lpadmin: Class name can only contain letters, numbers, and the underscore!\n", stderr);
+	      return (1);
+	    }
+
+            delete_printer_from_class(http, printer, pclass);
+	    break;
+
+        case 'u' : /* Allow/deny users */
+	    if (argv[i][2])
+	      val = argv[i] + 2;
+	    else
+	    {
+	      i ++;
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected name=value after \'-o\' option!\n", stderr);
+		return (1);
+	      }
+
+              val = argv[i];
+	    }
+
+            if (strncasecmp(val, "allow:", 6) == 0)
+	      num_options = cupsAddOption("requesting-user-name-allowed",
+	                                  val + 6, num_options, &options);
+            else if (strncasecmp(val, "deny:", 5) == 0)
+	      num_options = cupsAddOption("requesting-user-name-denied",
+	                                  val + 5, num_options, &options);
+            else
+	    {
+	      fprintf(stderr, "lpadmin: Unknown allow/deny option \"%s\"!\n",
+	              val);
+	      return (1);
 	    }
 	    break;
 
         case 'v' : /* Set the device-uri attribute */
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
 	    if (printer == NULL)
 	    {
 	      fputs("lpadmin: Unable to set the device URI:\n", stderr);
@@ -251,33 +458,48 @@ main(int  argc,		/* I - Number of command-line arguments */
 	    else
 	    {
 	      i ++;
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected device URI after \'-v\' option!\n", stderr);
+		return (1);
+	      }
+
 	      set_printer_device(http, printer, argv[i]);
 	    }
 	    break;
 
         case 'x' : /* Delete a printer */
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
 	    if (argv[i][2])
 	      printer = argv[i] + 2;
 	    else
 	    {
 	      i ++;
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected printer or class after \'-x\' option!\n", stderr);
+		return (1);
+	      }
+
 	      printer = argv[i];
 	    }
 
-	    if ((host = strchr(printer, '@')) != NULL)
+            if (!validate_name(printer))
 	    {
-	     /*
-	      * printer@host - reconnect to the named host...
-	      */
-
-	      httpClose(http);
-
-              *host++ = '\0';
-              if ((http = httpConnect(host, ippPort())) == NULL)
-	      {
-		perror("lpadmin: Unable to connect to server");
-		return (1);
-	      }
+	      fputs("lpadmin: Printer name can only contain letters, numbers, and the underscore!\n", stderr);
+	      return (1);
 	    }
 
             delete_printer(http, printer);
@@ -285,6 +507,17 @@ main(int  argc,		/* I - Number of command-line arguments */
 	    break;
 
         case 'D' : /* Set the printer-info attribute */
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
 	    if (printer == NULL)
 	    {
 	      fputs("lpadmin: Unable to set the printer description:\n", stderr);
@@ -297,11 +530,29 @@ main(int  argc,		/* I - Number of command-line arguments */
 	    else
 	    {
 	      i ++;
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected description after \'-D\' option!\n", stderr);
+		return (1);
+	      }
+
 	      set_printer_info(http, printer, argv[i]);
 	    }
 	    break;
 
         case 'L' : /* Set the printer-location attribute */
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
 	    if (printer == NULL)
 	    {
 	      fputs("lpadmin: Unable to set the printer location:\n", stderr);
@@ -314,11 +565,29 @@ main(int  argc,		/* I - Number of command-line arguments */
 	    else
 	    {
 	      i ++;
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected location after \'-L\' option!\n", stderr);
+		return (1);
+	      }
+
 	      set_printer_location(http, printer, argv[i]);
 	    }
 	    break;
 
         case 'P' : /* Use the specified PPD file */
+	    if (!http)
+	    {
+              http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+	      if (http == NULL)
+	      {
+		perror("lpadmin: Unable to connect to server");
+		return (1);
+	      }
+            }
+
 	    if (printer == NULL)
 	    {
 	      fputs("lpadmin: Unable to set the PPD file:\n", stderr);
@@ -331,6 +600,13 @@ main(int  argc,		/* I - Number of command-line arguments */
 	    else
 	    {
 	      i ++;
+
+	      if (i >= argc)
+	      {
+		fputs("lpadmin: Expected PPD after \'-P\' option!\n", stderr);
+		return (1);
+	      }
+
 	      set_printer_file(http, printer, argv[i]);
 	    }
 	    break;
@@ -345,6 +621,33 @@ main(int  argc,		/* I - Number of command-line arguments */
       return (1);
     }
 
+ /*
+  * Set options as needed...
+  */
+
+  if (num_options)
+  {
+    if (!http)
+    {
+      http = httpConnectEncrypt(cupsServer(), ippPort(), encryption);
+
+      if (http == NULL)
+      {
+	perror("lpadmin: Unable to connect to server");
+	return (1);
+      }
+    }
+
+    if (printer == NULL)
+    {
+      fputs("lpadmin: Unable to set the printer options:\n", stderr);
+      fputs("         You must specify a printer name first!\n", stderr);
+      return (1);
+    }
+
+    set_printer_options(http, printer, num_options, options);
+  }
+
   if (printer == NULL)
   {
     puts("Usage:");
@@ -353,11 +656,13 @@ main(int  argc,		/* I - Number of command-line arguments */
     puts("    lpadmin [-h server] -x destination");
     puts("    lpadmin [-h server] -p printer [-c add-class] [-i interface] [-m model]");
     puts("                       [-r remove-class] [-v device] [-D description]");
-    puts("                       [-P ppd-file]");
+    puts("                       [-P ppd-file] [-o name=value]");
+    puts("                       [-u allow:user,user] [-u deny:user,user]");
     puts("");
   }
 
-  httpClose(http);
+  if (http)
+    httpClose(http);
 
   return (0);
 }
@@ -381,7 +686,7 @@ add_printer_to_class(http_t *http,	/* I - Server connection */
   char		uri[HTTP_MAX_URI];	/* URI for printer/class */
 
 
-  DEBUG_printf(("add_printer_to_class(%08x, \"%s\", \"%s\")\n", http,
+  DEBUG_printf(("add_printer_to_class(%p, \"%s\", \"%s\")\n", http,
                 printer, pclass));
 
  /*
@@ -415,7 +720,7 @@ add_printer_to_class(http_t *http,	/* I - Server connection */
   * Do the request and get back a response...
   */
 
-  response = cupsDoRequest(http, request, "/classes/");
+  response = cupsDoRequest(http, request, "/");
 
  /*
   * Build a CUPS_ADD_CLASS request, which requires the following
@@ -480,7 +785,7 @@ add_printer_to_class(http_t *http,	/* I - Server connection */
     attr->values[i].string.text = strdup(uri);
   }
   else
-    ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_URI, "member-uris", NULL, uri);
+    attr = ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_URI, "member-uris", NULL, uri);
 
  /*
   * Then send the request...
@@ -489,9 +794,16 @@ add_printer_to_class(http_t *http,	/* I - Server connection */
   ippDelete(response);
 
   if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
-    fputs("lpadmin: Unable to add printer to class!\n", stderr);
+    fprintf(stderr, "lpadmin: add-class failed: %s\n",
+            ippErrorString(cupsLastError()));
   else
+  {
+    if (response->request.status.status_code > IPP_OK_CONFLICT)
+      fprintf(stderr, "lpadmin: add-class failed: %s\n",
+              ippErrorString(response->request.status.status_code));
+
     ippDelete(response);
+  }
 }
 
 
@@ -509,7 +821,7 @@ default_printer(http_t *http,		/* I - Server connection */
   char		uri[HTTP_MAX_URI];	/* URI for printer/class */
 
 
-  DEBUG_printf(("default_printer(%08x, \"%s\")\n", http, printer));
+  DEBUG_printf(("default_printer(%p, \"%s\")\n", http, printer));
 
  /*
   * Build a CUPS_SET_DEFAULT request, which requires the following
@@ -543,11 +855,13 @@ default_printer(http_t *http,		/* I - Server connection */
   */
 
   if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
-    fputs("lpadmin: Unable to set default destination!\n", stderr);
+    fprintf(stderr, "lpadmin: set-default failed: %s\n",
+            ippErrorString(cupsLastError()));
   else
   {
-    if (response->request.status.status_code == IPP_NOT_FOUND)
-      fprintf(stderr, "lpadmin: Destination %s does not exist!\n", printer);
+    if (response->request.status.status_code > IPP_OK_CONFLICT)
+      fprintf(stderr, "lpadmin: set-default failed: %s\n",
+              ippErrorString(response->request.status.status_code));
 
     ippDelete(response);
   }
@@ -568,7 +882,7 @@ delete_printer(http_t *http,		/* I - Server connection */
   char		uri[HTTP_MAX_URI];	/* URI for printer/class */
 
 
-  DEBUG_printf(("delete_printer(%08x, \"%s\")\n", http, printer));
+  DEBUG_printf(("delete_printer(%p, \"%s\")\n", http, printer));
 
  /*
   * Build a CUPS_DELETE_PRINTER request, which requires the following
@@ -602,11 +916,13 @@ delete_printer(http_t *http,		/* I - Server connection */
   */
 
   if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
-    fputs("lpadmin: Unable to delete printer!\n", stderr);
+    fprintf(stderr, "lpadmin: delete-printer failed: %s\n",
+            ippErrorString(cupsLastError()));
   else
   {
-    if (response->request.status.status_code == IPP_NOT_FOUND)
-      fprintf(stderr, "lpadmin: Destination %s does not exist!\n", printer);
+    if (response->request.status.status_code > IPP_OK_CONFLICT)
+      fprintf(stderr, "lpadmin: delete-printer failed: %s\n",
+              ippErrorString(response->request.status.status_code));
 
     ippDelete(response);
   }
@@ -631,7 +947,7 @@ delete_printer_from_class(http_t *http,		/* I - Server connection */
   char		uri[HTTP_MAX_URI];	/* URI for printer/class */
 
 
-  DEBUG_printf(("delete_printer_from_class(%08x, \"%s\", \"%s\")\n", http,
+  DEBUG_printf(("delete_printer_from_class(%p, \"%s\", \"%s\")\n", http,
                 printer, pclass));
 
  /*
@@ -771,9 +1087,16 @@ delete_printer_from_class(http_t *http,		/* I - Server connection */
   ippDelete(response);
 
   if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
-    fputs("lpadmin: Unable to remove printer from class!\n", stderr);
+    fprintf(stderr, "lpadmin: add/delete-class failed: %s\n",
+            ippErrorString(cupsLastError()));
   else
+  {
+    if (response->request.status.status_code > IPP_OK_CONFLICT)
+      fprintf(stderr, "lpadmin: add/delete-class failed: %s\n",
+              ippErrorString(response->request.status.status_code));
+
     ippDelete(response);
+  }
 }
 
 
@@ -791,7 +1114,7 @@ enable_printer(http_t *http,		/* I - Server connection */
   char		uri[HTTP_MAX_URI];	/* URI for printer/class */
 
 
-  DEBUG_printf(("enable_printer(%08x, \"%s\")\n", http, printer));
+  DEBUG_printf(("enable_printer(%p, \"%s\")\n", http, printer));
 
  /*
   * Build a CUPS_ADD_PRINTER request, which requires the following
@@ -832,17 +1155,67 @@ enable_printer(http_t *http,		/* I - Server connection */
   */
 
   if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
-    fputs("lpadmin: Unable to delete printer!\n", stderr);
+    fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+            ippErrorString(cupsLastError()));
   else
   {
-    if (response->request.status.status_code == IPP_NOT_FOUND)
-      fprintf(stderr, "lpadmin: Destination %s does not exist!\n", printer);
+    if (response->request.status.status_code > IPP_OK_CONFLICT)
+      fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+              ippErrorString(response->request.status.status_code));
 
     ippDelete(response);
   }
 }
 
 
+/*
+ * 'get_line()' - Get a line that is terminated by a LF, CR, or CR LF.
+ */
+
+static char *		/* O - Pointer to buf or NULL on EOF */
+get_line(char *buf,	/* I - Line buffer */
+         int  length,	/* I - Length of buffer */
+	 FILE *fp)	/* I - File to read from */
+{
+  char	*bufptr;	/* Pointer into buffer */
+  int	ch;		/* Character from file */
+
+
+  length --;
+  bufptr = buf;
+
+  while ((ch = getc(fp)) != EOF)
+  {
+    if (ch == '\n')
+      break;
+    else if (ch == '\r')
+    {
+     /*
+      * Look for LF...
+      */
+
+      ch = getc(fp);
+      if (ch != '\n' && ch != EOF)
+        ungetc(ch, fp);
+
+      break;
+    }
+
+    *bufptr++ = ch;
+    length --;
+    if (length == 0)
+      break;
+  }
+
+  *bufptr = '\0';
+
+  if (ch == EOF)
+    return (NULL);
+  else
+    return (buf);
+}
+
+
 /*
  * 'set_printer_device()' - Set the device-uri attribute.
  */
@@ -858,7 +1231,7 @@ set_printer_device(http_t *http,	/* I - Server connection */
   char		uri[HTTP_MAX_URI];	/* URI for printer/class */
 
 
-  DEBUG_printf(("set_printer_device(%08x, \"%s\", \"%s\")\n", http, printer,
+  DEBUG_printf(("set_printer_device(%p, \"%s\", \"%s\")\n", http, printer,
                 device));
 
  /*
@@ -911,9 +1284,16 @@ set_printer_device(http_t *http,	/* I - Server connection */
   */
 
   if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
-    fputs("lpadmin: Unable to set device-uri attribute!\n", stderr);
+    fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+            ippErrorString(cupsLastError()));
   else
+  {
+    if (response->request.status.status_code > IPP_OK_CONFLICT)
+      fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+              ippErrorString(response->request.status.status_code));
+
     ippDelete(response);
+  }
 }
 
 
@@ -932,13 +1312,13 @@ set_printer_file(http_t *http,		/* I - Server connection */
   char		uri[HTTP_MAX_URI];	/* URI for printer/class */
 #ifdef HAVE_LIBZ
   char		tempfile[1024];		/* Temporary filename */
-  FILE		*fp;			/* Temporary file */
+  int		fd;			/* Temporary file */
   gzFile	*gz;			/* GZIP'd file */
   char		buffer[8192];		/* Copy buffer */
   int		bytes;			/* Bytes in buffer */
 
 
-  DEBUG_printf(("set_printer_file(%08x, \"%s\", \"%s\")\n", http, printer,
+  DEBUG_printf(("set_printer_file(%p, \"%s\", \"%s\")\n", http, printer,
                 file));
 
  /*
@@ -952,7 +1332,7 @@ set_printer_file(http_t *http,		/* I - Server connection */
     * Yes, the file is compressed; uncompress to a temp file...
     */
 
-    if ((fp = fopen(cupsTempFile(tempfile, sizeof(tempfile)), "wb")) == NULL)
+    if ((fd = cupsTempFd(tempfile, sizeof(tempfile))) < 0)
     {
       perror("lpadmin: Unable to create temporary file");
       return;
@@ -961,15 +1341,15 @@ set_printer_file(http_t *http,		/* I - Server connection */
     if ((gz = gzopen(file, "rb")) == NULL)
     {
       perror("lpadmin: Unable to open file");
-      fclose(fp);
+      close(fd);
       unlink(tempfile);
       return;
     }
 
     while ((bytes = gzread(gz, buffer, sizeof(buffer))) > 0)
-      fwrite(buffer, bytes, 1, fp);
+      write(fd, buffer, bytes);
 
-    fclose(fp);
+    close(fd);
     gzclose(gz);
 
     file = tempfile;
@@ -1008,9 +1388,16 @@ set_printer_file(http_t *http,		/* I - Server connection */
   */
 
   if ((response = cupsDoFileRequest(http, request, "/admin/", file)) == NULL)
-    fputs("lpadmin: Unable to set interface script or PPD file!\n", stderr);
+    fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+            ippErrorString(cupsLastError()));
   else
+  {
+    if (response->request.status.status_code > IPP_OK_CONFLICT)
+      fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+              ippErrorString(response->request.status.status_code));
+
     ippDelete(response);
+  }
 
 #ifdef HAVE_LIBZ
  /*
@@ -1038,7 +1425,7 @@ set_printer_info(http_t *http,		/* I - Server connection */
   char		uri[HTTP_MAX_URI];	/* URI for printer/class */
 
 
-  DEBUG_printf(("set_printer_info(%08x, \"%s\", \"%s\")\n", http, printer,
+  DEBUG_printf(("set_printer_info(%p, \"%s\", \"%s\")\n", http, printer,
                 info));
 
  /*
@@ -1080,9 +1467,16 @@ set_printer_info(http_t *http,		/* I - Server connection */
   */
 
   if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
-    fputs("lpadmin: Unable to set printer-info attribute!\n", stderr);
+    fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+            ippErrorString(cupsLastError()));
   else
+  {
+    if (response->request.status.status_code > IPP_OK_CONFLICT)
+      fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+              ippErrorString(response->request.status.status_code));
+
     ippDelete(response);
+  }
 }
 
 
@@ -1101,7 +1495,7 @@ set_printer_location(http_t *http,	/* I - Server connection */
   char		uri[HTTP_MAX_URI];	/* URI for printer/class */
 
 
-  DEBUG_printf(("set_printer_location(%08x, \"%s\", \"%s\")\n", http, printer,
+  DEBUG_printf(("set_printer_location(%p, \"%s\", \"%s\")\n", http, printer,
                 location));
 
  /*
@@ -1143,12 +1537,320 @@ set_printer_location(http_t *http,	/* I - Server connection */
   */
 
   if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
-    fputs("lpadmin: Unable to set printer-location attribute!\n", stderr);
+    fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+            ippErrorString(cupsLastError()));
   else
+  {
+    if (response->request.status.status_code > IPP_OK_CONFLICT)
+      fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+              ippErrorString(response->request.status.status_code));
+
     ippDelete(response);
+  }
 }
 
 
 /*
- * End of "$Id: lpadmin.c,v 1.10 1999/10/10 15:41:15 mike Exp $".
+ * 'set_printer_model()' - Set the driver model file.
+ */
+
+static void
+set_printer_model(http_t *http,		/* I - Server connection */
+                  char   *printer,	/* I - Printer */
+		  char   *model)	/* I - Driver model file */
+{
+  ipp_t		*request,		/* IPP Request */
+		*response;		/* IPP Response */
+  cups_lang_t	*language;		/* Default language */
+  char		uri[HTTP_MAX_URI];	/* URI for printer/class */
+
+
+ /*
+  * Build a CUPS_ADD_PRINTER request, which requires the following
+  * attributes:
+  *
+  *    attributes-charset
+  *    attributes-natural-language
+  *    printer-uri
+  *    ppd-name
+  */
+
+  snprintf(uri, sizeof(uri), "ipp://localhost/printers/%s", printer);
+
+  request = ippNew();
+
+  request->request.op.operation_id = CUPS_ADD_PRINTER;
+  request->request.op.request_id   = 1;
+
+  language = cupsLangDefault();
+
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
+               "attributes-charset", NULL, cupsLangEncoding(language));
+
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
+               "attributes-natural-language", NULL, language->language);
+
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
+               "printer-uri", NULL, uri);
+
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_NAME,
+               "ppd-name", NULL, model);
+
+ /*
+  * Do the request and get back a response...
+  */
+
+  if ((response = cupsDoRequest(http, request, "/admin/")) == NULL)
+    fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+            ippErrorString(cupsLastError()));
+  else
+  {
+    if (response->request.status.status_code > IPP_OK_CONFLICT)
+      fprintf(stderr, "lpadmin: add-printer failed: %s\n",
+              ippErrorString(response->request.status.status_code));
+
+    ippDelete(response);
+  }
+}
+
+
+/*
+ * 'set_printer_options()' - Set the printer options.
+ */
+
+static void
+set_printer_options(http_t        *http,	/* I - Server connection */
+                    char          *printer,	/* I - Printer */
+		    int           num_options,	/* I - Number of options */
+		    cups_option_t *options)	/* I - Options */
+{
+  ipp_t		*request,		/* IPP Request */
+		*response;		/* IPP Response */
+  ipp_attribute_t *attr;		/* IPP attribute */
+  cups_lang_t	*language;		/* Default language */
+  ipp_op_t	op;			/* Operation to perform */
+  const char	*val,			/* Option value */
+		*ppdfile;		/* PPD filename */
+  char		uri[HTTP_MAX_URI],	/* URI for printer/class */
+		line[1024],		/* Line from PPD file */
+		keyword[1024],		/* Keyword from Default line */
+		*keyptr,		/* Pointer into keyword... */
+		tempfile[1024];		/* Temporary filename */
+  FILE		*in,			/* PPD file */
+		*out;			/* Temporary file */
+  int		outfd;			/* Temporary file descriptor */
+
+
+  DEBUG_printf(("set_printer_options(%p, \"%s\", %d, %p)\n", http, printer,
+                num_options, options));
+
+  language = cupsLangDefault();
+
+  snprintf(uri, sizeof(uri), "ipp://localhost/printers/%s", printer);
+
+ /*
+  * Build a GET_PRINTER_ATTRIBUTES request, which requires the following
+  * attributes:
+  *
+  *    attributes-charset
+  *    attributes-natural-language
+  *    printer-uri
+  *    requested-attributes
+  */
+
+  request = ippNew();
+
+  request->request.op.operation_id = IPP_GET_PRINTER_ATTRIBUTES;
+  request->request.op.request_id   = 1;
+
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
+               "attributes-charset", NULL, cupsLangEncoding(language));
+
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
+               "attributes-natural-language", NULL, language->language);
+
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
+               "printer-uri", NULL, uri);
+
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD,
+               "requested-attributes", NULL, "printer-type");
+
+ /*
+  * Do the request...
+  */
+
+  op = CUPS_ADD_PRINTER;
+
+  if ((response = cupsDoRequest(http, request, "/")) != NULL)
+  {
+   /*
+    * See what kind of printer or class it is...
+    */
+
+    if ((attr = ippFindAttribute(response, "printer-type", IPP_TAG_ENUM)) != NULL)
+    {
+      if (attr->values[0].integer & (CUPS_PRINTER_CLASS | CUPS_PRINTER_IMPLICIT))
+        op = CUPS_ADD_CLASS;
+    }
+
+    ippDelete(response);
+  }
+
+ /*
+  * Build a CUPS_ADD_PRINTER or CUPS_ADD_CLASS request, which requires
+  * the following attributes:
+  *
+  *    attributes-charset
+  *    attributes-natural-language
+  *    printer-uri
+  *    other options
+  */
+
+  request = ippNew();
+
+  request->request.op.operation_id = op;
+  request->request.op.request_id   = 1;
+
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
+               "attributes-charset", NULL, cupsLangEncoding(language));
+
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
+               "attributes-natural-language", NULL, language->language);
+
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
+               "printer-uri", NULL, uri);
+
+ /*
+  * Add the options...
+  */
+
+  cupsEncodeOptions(request, num_options, options);
+
+  if (op == CUPS_ADD_PRINTER)
+    ppdfile = cupsGetPPD(printer);
+  else
+    ppdfile = NULL;
+
+  if (ppdfile != NULL)
+  {
+   /*
+    * Set default options in the PPD file...
+    */
+
+    outfd = cupsTempFd(tempfile, sizeof(tempfile));
+
+    in  = fopen(ppdfile, "rb");
+    out = fdopen(outfd, "wb");
+
+    while (get_line(line, sizeof(line), in) != NULL)
+    {
+      if (strncmp(line, "*Default", 8) != 0)
+        fprintf(out, "%s\n", line);
+      else
+      {
+       /*
+        * Get default option name...
+	*/
+
+        strncpy(keyword, line + 8, sizeof(keyword) - 1);
+	keyword[sizeof(keyword) - 1] = '\0';
+
+	for (keyptr = keyword; *keyptr; keyptr ++)
+	  if (*keyptr == ':' || isspace(*keyptr))
+	    break;
+
+        *keyptr = '\0';
+
+        if (strcmp(keyword, "PageRegion") == 0)
+	  val = cupsGetOption("PageSize", num_options, options);
+	else
+	  val = cupsGetOption(keyword, num_options, options);
+
+        if (val != NULL)
+	  fprintf(out, "*Default%s: %s\n", keyword, val);
+	else
+	  fprintf(out, "%s\n", line);
+      }
+    }
+
+    fclose(in);
+    fclose(out);
+    close(outfd);
+
+   /*
+    * Do the request...
+    */
+
+    response = cupsDoFileRequest(http, request, "/admin/", tempfile);
+
+   /*
+    * Clean up temp files... (TODO: catch signals in case we CTRL-C during
+    * lpadmin)
+    */
+
+    unlink(ppdfile);
+    unlink(tempfile);
+  }
+  else
+  {
+   /*
+    * No PPD file - just set the options...
+    */
+
+    response = cupsDoRequest(http, request, "/admin/");
+  }
+
+ /*
+  * Check the response...
+  */
+
+  if (response == NULL)
+    fprintf(stderr, "lpadmin: %s failed: %s\n",
+            op == CUPS_ADD_PRINTER ? "add-printer" : "add-class",
+            ippErrorString(cupsLastError()));
+  else
+  {
+    if (response->request.status.status_code > IPP_OK_CONFLICT)
+      fprintf(stderr, "lpadmin: %s failed: %s\n",
+              op == CUPS_ADD_PRINTER ? "add-printer" : "add-class",
+              ippErrorString(response->request.status.status_code));
+
+    ippDelete(response);
+  }
+}
+
+
+/*
+ * 'validate_name()' - Make sure the printer name only contains letters,
+ *                     numbers, and the underscore...
+ */
+
+static int			/* O - 0 if name is no good, 1 if name is good */
+validate_name(const char *name)	/* I - Name to check */
+{
+ /*
+  * Don't allow names to start with a digit...
+  */
+
+  if (isdigit(*name))
+    return (0);
+
+ /*
+  * Scan the whole name...
+  */
+
+  while (*name)
+    if (*name == '@')
+      return (1);
+    else if (!isalnum(*name) && *name != '_')
+      return (0);
+    else
+      name ++;
+
+  return (1);
+}
+
+
+/*
+ * End of "$Id: lpadmin.c,v 1.28 2002/01/02 17:59:19 mike Exp $".
  */
